from typing import Union

import pygef.plot_utils as plot
from pygef.base import Base
from pygef.broxml import _BroXmlBore
from pygef.gef import _GefBore


class Bore(Base):
    """
    ** Bore attributes:**
        type: str
            Type of the gef file
        project_id: str
            Project id
        x: float
            X coordinate respect to the coordinate system
        y: float
            Y coordinate respect to the coordinate system
        zid: float
            Z coordinate respect to the height system
        height_system: float
            Type of coordinate system, 31000 is NAP
        file_date: datatime.datetime
            Start date time
        test_id: str
            Identifying name of gef file.
        s: str
            String version of gef file.
        df: polars.DataFrame
            DataFrame containing the columns: [
                                                "depth_top",
                                                "depth_bottom",
                                                "soil_code",
                                                "gravel_component",
                                                "sand_component",
                                                "clay_component",
                                                "loam_component",
                                                "peat_component",
                                                "silt_component",
                                                "remarks",
                                            ]
    """

    def __init__(self, path=None, content: dict = None):
        """
        Bore class.

        Parameters
        ----------
        path:
            Path to the file.
        content: dict
            Dictionary with keys: ["string", "file_type"]
                -string: str
                    String version of the file.
                -file_type:
                    One of [gef, xml]
        """
        super().__init__()

        parsed: Union[_BroXmlBore, _GefBore]

        self.nen_version = None

        if content is not None:
            assert (
                content["file_type"] == "gef" or content["file_type"] == "xml"
            ), f"file_type can be only one of [gef, xml] "
            assert content["string"] is not None, "content['string'] must be specified"
            if content["file_type"] == "gef":
                parsed = _GefBore(string=content["string"])
            elif content["file_type"] == "xml":
                parsed = _BroXmlBore(string=content["string"])

        elif path is not None:
            if path.lower().endswith("gef"):
                parsed = _GefBore(path)
            elif path.lower().endswith("xml"):
                parsed = _BroXmlBore(path)
        else:
            raise ValueError("One of [path, (string, file_type)] should be not None.")

        self.__dict__.update(parsed.__dict__)

    def plot(self, figsize=(11, 8), show=False, dpi=100):
        """
        Plot plot of soil components over the depth.

        Parameters
        ----------
        show: bool
            If True the plot is showed, else the matplotlib.pytplot.figure is returned
        figsize: tuple
            Figsize of the plot, default (11, 8).
        dpi: int
            Dpi figure
        Returns
        -------
        matplotlib.pyplot.figure
        """
        if self.nen_version == "NEN 5104":
            return plot.plot_bore(self.df, figsize=figsize, show=show, dpi=dpi)
        elif self.nen_version == "NEN-EN-ISO 14688":
            return plot.plot_bore(self.df, figsize=figsize, show=show, dpi=dpi)
        else:
            raise ValueError(
                "Supported NEN version are only: 'NEN 5104' and 'NEN-EN-ISO 14688'"
            )



from pygef.been_jefferies.util import iterate_been_jeffrey


def classify(
    df,
    water_level_and_zid_NAP=None,
    water_level_wrt_depth=None,
    area_quotient_cone_tip=None,
    pre_excavated_depth=None,
):
    """
    Classify function for Been & Jefferies.

    :param df: (DataFrame) Original dataframe.
    :param water_level_and_zid_NAP: (float)
    :param water_level_wrt_depth: (float)
    :param area_quotient_cone_tip: (float)
    :param pre_excavated_depth: (float)
    :return:
    """
    if water_level_and_zid_NAP["water_level_NAP"] is not None:
        water_level = (
            water_level_and_zid_NAP["zid"] - water_level_and_zid_NAP["water_level_NAP"]
        )
    elif water_level_wrt_depth is not None:
        water_level = water_level_wrt_depth
    else:
        raise ValueError(
            "You should specify one of water_level_NAP / water_level_wrt_depth"
        )
    return iterate_been_jeffrey(
        df,
        water_level,
        area_quotient_cone_tip=area_quotient_cone_tip,
        pre_excavated_depth=pre_excavated_depth,
    )



import numpy as np


def delta_depth(df, pre_excavated_depth=None):
    """
    Take pre-excavated depth into account.

    This function corrects the depth column and adds a column (np.diff(depth))

    :param df: (DataFrame) with depth column
    :param pre_excavated_depth: (flt)
    :return: (DataFrame) [depth, delta_depth]
    """
    df["delta_depth"] = np.r_[np.array([df["depth"][0]]), np.diff(df["depth"])]

    return df


def soil_pressure(df):
    df["soil_pressure"] = (df["gamma"] * df["delta_depth"]).cumsum()

    return df


def water_pressure(df, water_level):
    df["water_pressure"] = df["depth"].apply(lambda depth: (depth - water_level) * 9.81)
    df[df["water_pressure"] < 0, "water_pressure"] = 0.0

    return df


def effective_soil_pressure(df):
    df["effective_soil_pressure"] = df["soil_pressure"] - df["water_pressure"]

    return df


def qt(df, area_quotient_cone_tip=None):
    if "u2" in df.columns and area_quotient_cone_tip is not None:
        df["qt"] = df["qc"] + df["u2"].apply(
            lambda u2: u2 * (1.0 - area_quotient_cone_tip)
        )
    else:
        df["qt"] = df["qc"]

    return df


def normalized_cone_resistance(df):
    df["normalized_cone_resistance"] = (df["qt"] - df["soil_pressure"]) / df[
        "effective_soil_pressure"
    ]
    df[df["normalized_cone_resistance"] < 0, "normalized_cone_resistance"] = 1.0

    return df


def normalized_friction_ratio(df):
    if "fs" in df.columns:
        fs = df["fs"]
    else:
        fs = df["friction_number"] * df["qc"] / 100.0

    df["normalized_friction_ratio"] = (fs / (df["qt"] - df["soil_pressure"])) * 100

    df[df["normalized_friction_ratio"] < 0, "normalized_friction_ratio"] = 0.1

    return df



from copy import copy

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import polars as pl

colours_robertson = {
    "Peat": "#a76b29",
    "Clays - silty clay to clay": "#578E57",
    "Silt mixtures - clayey silt to silty clay": "#0078C1",
    "Sand mixtures - silty sand to sandy silt": "#DBAD4B",
    "Sands - clean sand to silty sand": "#e5c581",
    "Gravelly sand to dense sand": "#708090",
    None: "black",
}

colours_been_jefferies = {
    "Peat": "#a76b29",
    "Clays": "#578E57",
    "Clayey silt to silty clay": "#0078C1",
    "Silty sand to sandy silt": "#DBAD4B",
    "Sands: clean sand to silty": "#e5c581",
    "Gravelly sands": "#708090",
    None: "black",
}


def num_columns(classification, df_group):
    """
    Get the columns number to plot.

    :param classification: (str) classification
    :param df_group: grouped dataframe.
    :return: number of columns to plot
    """
    if classification is None:
        return 2
    if df_group is None and classification is not None:
        return 4
    else:
        return 5


def plot_cpt(
    df, df_group, classification, show, figsize, grid_step_x, colors, dpi, z_NAP
):
    """
    Main function to plot qc, Fr and soil classification.

    :param df: Complete df.
    :param df_group: Grouped df.
    :param show: If show is True the figure is shown.
    :param figsize: Figure size (x, y)
    :return:
    """
    title = None
    title_group = None

    if classification is not None:
        df, title = assign_color(df, classification, colors)
    if df_group is not None:
        df_group = copy(df_group)
        df_group = df_group.rename({"layer": "soil_type"})
        df_group, title_group = assign_color(df_group, classification, colors=colors)
        if colors is None:
            title_group = "Filtered"
        else:
            title_group = "User defined filter"
    if z_NAP:
        depth_max = df["elevation_with_respect_to_nap"].min()
        depth_min = df["elevation_with_respect_to_nap"].max()
    else:
        depth_max = df["depth"].max()
        depth_min = df["depth"].min()

    fig = plt.figure(figsize=figsize, dpi=dpi)
    n = 0
    num_col = num_columns(classification, df_group)
    for c, unit in zip(["qc", "friction_number"], ["[MPa]", "[%]"]):
        n += 1
        ax = fig.add_subplot(1, num_col, n)
        if z_NAP:
            plt.plot(df[c], df["elevation_with_respect_to_nap"], "C0")
            if n == 1:
                ax.set_ylabel("Z NAP [m]")
        else:
            plt.plot(df[c], df["depth"], "C0")
            if n == 1:
                ax.set_ylabel("Z [m]")
        if n > 1:
            # keep grid, but no labels
            ax = plt.gca()
            labels = [item.get_text() for item in ax.get_yticklabels()]
            empty_string_labels = [""] * len(labels)
            ax.set_yticklabels(empty_string_labels)
            plt.xlim([0, min(15, df[c].max() + 1)])
        plt.xlim(0, df[c].max() * 1.05)
        ax.set_xlabel(f"{c} {unit}")

        plt.grid()

        # custom x grid
        if grid_step_x is not None:
            ax.set_xticks(np.arange(0, df[c].max() + grid_step_x, grid_step_x))
        plt.ylim(depth_max, depth_min)

    if classification is not None:
        fig = add_plot_classification(
            fig, df, depth_max, depth_min, title, num_col, z_NAP=z_NAP
        )
    if df_group is not None:
        fig = add_grouped_classification(
            fig, df_group, depth_max, depth_min, title_group, num_col, z_NAP=z_NAP
        )

    if classification is not None:
        # custom legend
        legend_dict = get_legend(classification, colors=colors)
        patch_list = []
        for key in legend_dict:
            data_key = mpatches.Patch(color=legend_dict[key], label=key)
            patch_list.append(data_key)

        plt.legend(handles=patch_list, bbox_to_anchor=(1, 1), loc="upper left")
    if show:
        plt.show()
    else:
        return fig


def assign_color(df, classification, colors):
    """
    Add to the dataframe the column 'colour' based on the chosen classification.

    :param df: original dataframe.
    :param classification: Chosen classification.
    :param colors: (Dictionary) Dictionary with the user color associated to each layer.
    :return:
    """
    if colors is None:
        if classification == "robertson":
            df = df.with_column(
                pl.col("soil_type")
                .apply(lambda row: colours_robertson[row])
                .alias("colour")
            )

            return (
                df,
                "Robertson",
            )
        elif classification == "been_jefferies":
            df = df.with_column(
                pl.col("soil_type")
                .apply(lambda row: colours_been_jefferies[row])
                .alias("colour")
            )

            return (
                df,
                "Been Jefferies",
            )
    else:
        df = df.with_column(
            pl.col("soil_type").apply(lambda row: colors[row]).alias("colour")
        )

        return (
            df,
            "User defined",
        )


def add_plot_classification(fig, df, depth_max, depth_min, title, num_col, z_NAP):
    """
    Add to the plot the selected classification.

    :param fig: Original figure.
    :param depth_max: Maximum depth.
    :param depth_min: Minimum depth.
    :return:
    """
    ax = fig.add_subplot(1, num_col, 3)
    df[df["soil_type"].is_null(), "soil_type"] = "UNKNOWN"
    for st in df["soil_type"].unique():
        partial_df = df[df["soil_type"] == st]
        if z_NAP:
            plt.hlines(
                y=partial_df["elevation_with_respect_to_nap"],
                xmin=0,
                xmax=1,
                colors=partial_df["colour"],
                label=st,
            )
        else:
            plt.hlines(
                y=partial_df["depth"],
                xmin=0,
                xmax=1,
                colors=partial_df["colour"],
                label=st,
            )
    ax.set_xticks([])
    ax.set_title(f"{title}")
    plt.ylim(depth_max, depth_min)
    return fig


def add_grouped_classification(
    fig, df_group, depth_max, depth_min, title_group, num_col, z_NAP
):
    """
    Add to the plot the selected classification.

    :param fig: Original figure.
    :param depth_max: Maximum depth.
    :param depth_min: Minimum depth.
    :return: Complete figure.
    """
    ax = fig.add_subplot(1, num_col, 4)
    df = df_group
    for i, layer in enumerate(df["soil_type"]):
        if z_NAP:
            plt.barh(
                y=df["z_centr_nap"][i],
                height=df["thickness"][i],
                width=5,
                color=df["colour"][i],
                label=layer,
            )
        else:
            plt.barh(
                y=df["z_centr"][i],
                height=df["thickness"][i],
                width=5,
                color=df["colour"][i],
                label=layer,
            )
    ax.set_xticks([])
    ax.set_title(f"{title_group} classification")
    plt.ylim(depth_max, depth_min)
    return fig


def plot_merged_cpt_bore(df, figsize=None, show=True):
    fig = plt.figure(figsize=figsize)
    subplot_val = 131
    plt.subplot(subplot_val)
    plt.plot(df["qc"], -df["depth"])

    subplot_val += 1
    plt.subplot(subplot_val)
    plt.plot(df["friction_number"], -df["depth"])

    subplot_val += 1
    plt.subplot(subplot_val)
    if "silt_component" in df.columns:
        df = df.copy()
        df["loam_component"] += df["silt_component"]
    v = df[
        [
            "gravel_component",
            "sand_component",
            "loam_component",
            "clay_component",
            "peat_component",
        ]
    ]

    c = ["#a76b29", "#578E57", "#0078C1", "#DBAD4B", "#708090"]
    for i in range(5):
        plt.fill_betweenx(
            -df["depth"],
            np.zeros(v.shape[0]),
            np.cumsum(v, axis=1)[:, -(i + 1)],
            color=c[i],
        )

    if show:
        plt.show()
    return fig


def plot_bore(df, figsize=(11, 8), show=True, dpi=100):
    fig = plt.figure(figsize=figsize, dpi=dpi)

    v = df[
        [
            "gravel_component",
            "sand_component",
            "loam_component",
            "clay_component",
            "peat_component",
        ]
    ]

    # TODO: replace with proper polars code
    vp = v.to_pandas()
    vp.iloc[:, 2] += df.to_pandas()["silt_component"]

    # TODO: what does this do?
    # vp[vp.sum(1) < 0] = np.nan

    c = ["#a76b29", "#578E57", "#0078C1", "#DBAD4B", "#708090"]

    for i in range(5):
        plt.fill_betweenx(
            -np.repeat(df["depth_top"], 2),
            np.zeros(vp.shape[0] * 2),
            np.roll(np.repeat(np.cumsum(vp, axis=1).iloc[:, -(i + 1)], 2), 1),
            color=c[i],
        )

    legend_dict = {
        "Gravel": "#708090",
        "Sand": "#DBAD4B",
        "Loam": "#0078C1",
        "Clay": "#578E57",
        "Peat": "#a76b29",
    }
    patch_list = []
    for key in legend_dict:
        data_key = mpatches.Patch(color=legend_dict[key], label=key)
        patch_list.append(data_key)

    plt.legend(handles=patch_list, bbox_to_anchor=(1, 1), loc="upper left")

    if show:
        plt.show()
    return fig


def get_legend(classification, colors):
    if colors is not None:
        return colors
    elif classification == "robertson":
        return colours_robertson
    elif classification == "been_jefferies":
        return colours_been_jefferies



from pygef.robertson.util import iterate_robertson


def classify(
    df,
    water_level_and_zid_NAP=None,
    water_level_wrt_depth=None,
    new=True,
    area_quotient_cone_tip=None,
    pre_excavated_depth=None,
    p_a=0.1,
):
    """
    Classify function for Rpbertson.

    :param df: (DataFrame) Original dataframe.
    :param water_level_NAP: (dict) {
                                        water_level: <water level wrt NAP,
                                        zid: <zid>
                                    }
    :param new: (bool): Old or New implementation of Robertson.
    :param area_quotient_cone_tip: (flt)
    :param pre_excavated_depth: (flt)
    :param p_a: (flt) Atmospheric pressure at ground level in MPA.
    :return: (DataFrame) containing classification and IC values.
    """

    if water_level_and_zid_NAP["water_level_NAP"] is not None:
        water_level = (
            water_level_and_zid_NAP["zid"] - water_level_and_zid_NAP["water_level_NAP"]
        )
    elif water_level_wrt_depth is not None:
        water_level = df["depth"].min() - water_level_wrt_depth
    else:
        raise ValueError(
            "You should specify one of water_level_NAP / water_level_wrt_depth"
        )

    return iterate_robertson(
        df,
        water_level,
        new=new,
        area_quotient_cone_tip=area_quotient_cone_tip,
        pre_excavated_depth=pre_excavated_depth,
        p_a=p_a,
    )



from pygef._version import __version__
from pygef.bore import Bore
from pygef.cpt import Cpt
from pygef.plot_utils import plot_merged_cpt_bore
from pygef.utils import depth_to_nap, join_gef, nap_to_depth



import logging
import re
from datetime import date

import numpy as np

logger = logging.getLogger(__name__)


def cast_string(f, s):
    """
    Generic function that casts a string.

    :param f: (function) Any casting function.
    :param s: (str) String to cast.
    :return: The casted object
    """
    try:
        return f(s)
    except ValueError as e:
        return None


def first_header_value(headers, name, index=0, cast=None):
    """
    Get the first matching line with a header.

    Throws an error when the header exists but doesn't have the value at the
    passed index.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :param name:(str) Header name.
    :param index:(int) Index of the value.
    :param cast: (function) Cast to the specified type.
    :return: (str) The header value.
    """
    if name in headers:
        result = headers[name][0][index]
        if cast:
            return cast(result)
        else:
            return result


def parse_regex_cast(regex_string, s, f, group_number):
    """
    Function that searches a regex match and casts the result.

    :param regex_string: (str) Regex pattern.
    :param s: (str) String to search for regex pattern.
    :param f: (function) To apply to the regex match.
    :param group_number: (int) Which group number to query when there is a match.
    :return: Casted result.
    """
    g = re.search(regex_string, s)
    if g:
        return cast_string(f, g.group(group_number))
    else:
        return None


def parse_end_of_header(s):
    """
    Function that parses the end of the header.

    :param s:(str) String to search for regex pattern.
    :return:(str) End of the header.
    """
    return parse_regex_cast(r"(#EOH[=\s+]+)", s, str, 1)


def parse_column_void(headers):
    """
    Function that parses the column void.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return:([str]) List of all the possible column void.
    """
    if isinstance(headers, dict):
        # Return a list of all the second float values of all COLUMN_VOID lines
        if "COLUMNVOID" in headers:
            return list(map(lambda values: float(values[1]), headers["COLUMNVOID"]))
    else:
        column_void = None
        g = re.findall(r"#COLUMNVOID[=\s+]+\d[,\s+]+([\d-]+\.?\d*)", headers)
        if g:
            column_void = list(
                map(
                    float,
                    re.findall(r"#COLUMNVOID[=\s+]+\d[,\s+]+([\d-]+\.?\d*)", headers),
                )
            )
            return column_void

    # Standard value, if some gef test_files the column void is not specified but used anyway
    return -9999


def parse_measurement_var_as_float(headers, var_number):
    """
    Function that returns a measurement variable as a float.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :param var_number: (int) Variable number.
    :return: Variable value.
    """
    # Convert the variable number to a string so it's cheaper to compare
    var_number_str = str(var_number)

    try:
        if isinstance(headers, dict):
            # Loop over all headers to find the right number
            if "MEASUREMENTVAR" in headers:
                for values in headers["MEASUREMENTVAR"]:
                    if values[0] == var_number_str:
                        return float(values[1])
        else:
            # Find all '#MEASUREMENTVAR= **,' strings first
            for match in re.finditer(
                r"#MEASUREMENTVAR[=\s+]+(\d+)[, ]+([\d-]+\.?\d*)", headers
            ):
                # The first group is the variable number
                if match.group(1) == var_number_str:
                    # The second group is the actual value
                    return cast_string(float, match.group(2))

    except ValueError:
        pass


def parse_cone_id(headers):
    """
    Function that returns the cone id specified in GEF-CPT file.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: cone id.
    """
    if isinstance(headers, dict):
        return first_header_value(headers, "MEASUREMENTTEXT", index=1)
    else:
        try:
            return parse_regex_cast(
                r"#MEASUREMENTTEXT[=\s+]+4[, ]+([\w.-]+)", headers, str, 1
            )
        except ValueError:
            return None


def parse_cpt_class(headers):
    """
    Function that returns the class of the cpt as an int.
    The word class or klasse has to be defined in the parsed string.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: Cpt class.
    """
    if isinstance(headers, dict):
        all_definition = first_header_value(headers, "MEASUREMENTTEXT", index=1)
    else:
        all_definition = parse_regex_cast(
            r"#MEASUREMENTTEXT[=\s+]+6[, ](.*)", headers, str, 1
        )

    if all_definition is not None:
        return parse_regex_cast(
            r"^.*?(klasse|class).*?(\d{1})", all_definition.lower(), int, 2
        )


def parse_project_type(headers, gef_type):
    """
    Function that returns the project type as an int.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :param gef_type: (str) String from which the gef type is given.
    :return: Project type number.
    """
    if isinstance(headers, dict):
        if gef_type == "cpt":
            try:
                # Try to get the second value first
                return first_header_value(headers, "PROJECTID", index=1)
            except Exception:
                # If that fails get the first one
                return first_header_value(headers, "PROJECTID")
        elif gef_type == "bore":
            return first_header_value(headers, "PROJECTID")
    else:
        if gef_type == "cpt":
            return parse_regex_cast(r"PROJECTID[\s=a-zA-Z,]*([\w-]+)", headers, str, 1)
        elif gef_type == "bore":
            return parse_regex_cast(r"#PROJECTID+[^a-zA-Z]+([\w-]+)", headers, str, 1)


def parse_zid_as_float(headers):
    """
    Function that returns the zid as a float.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return:(float) ZID number.
    """
    if isinstance(headers, dict):
        if "ZID" in headers:
            return first_header_value(headers, "ZID", index=1, cast=float)
    else:
        return parse_regex_cast(
            r"#ZID[=\s+]+[^,]*[,\s+]+([^?!,$|\s$]+)", headers, float, 1
        )


def parse_height_system(headers):
    """
    Function that returns the zid as a float.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return:(float) ZID number.
    """
    if isinstance(headers, dict):
        return first_header_value(headers, "ZID", cast=float)
    else:
        return parse_regex_cast(r"#ZID[=\s+]+([^,]*)", headers, float, 1)


def parse_xid_as_float(headers):
    """
    Function that returns the x coordinate as a float.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return:(float) x coordinate value.
    """
    if isinstance(headers, dict):
        return first_header_value(headers, "XYID", index=1, cast=float)
    else:
        return parse_regex_cast(
            r"#XYID[=\s+]*.*?,\s*(\d*\s*(\.|\d)*),\s*(\d*(\.|\d)*)", headers, float, 1
        )


def parse_yid_as_float(headers):
    """
    Function that returns the y coordinate as a float.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return:(float) y coordinate value.
    """
    if isinstance(headers, dict):
        return first_header_value(headers, "XYID", index=2, cast=float)
    else:
        return parse_regex_cast(
            r"#XYID[=\s+]*.*?,\s*(\d*\s*(\.|\d)*),\s*(\d*(\.|\d)*)", headers, float, 3
        )


def parse_gef_type(headers):
    """
    Function that returns the gef type.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: (str) Gef type.
    """
    proc_code = ""
    if isinstance(headers, dict):
        if "REPORTCODE" in headers:
            proc_code = first_header_value(headers, "REPORTCODE").lower()
        elif "PROCEDURECODE" in headers:
            proc_code = first_header_value(headers, "PROCEDURECODE").lower()
        else:
            return None
    else:
        if parse_regex_cast(
            r"#PROCEDURECODE[^a-zA-Z]+([\w-]+)", headers, lambda x: x.lower(), 1
        ) and parse_regex_cast(
            r"#REPORTCODE[^a-zA-Z]+([\w-]+)", headers, lambda x: x.lower(), 1
        ):
            proc_code = parse_regex_cast(
                r"#REPORTCODE[^a-zA-Z]+([\w-]+)", headers, lambda x: x.lower(), 1
            )
        else:
            proc_code = parse_regex_cast(
                r"#(REPORTCODE|PROCEDURECODE)[^a-zA-Z]+([\w-]+)",
                headers,
                lambda x: x.lower(),
                2,
            )

    if "cpt" in proc_code or "dis" in proc_code:
        return "cpt"
    elif "bore" in proc_code and not "borehole" in proc_code:
        return "bore"
    elif "borehole" in proc_code:
        return "borehole-report"


def parse_file_date(headers):
    """
    Fuction to parse the file date.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: File date.
    """
    year = None
    month = None
    day = None

    if isinstance(headers, dict):
        if "FILEDATE" in headers:
            year = first_header_value(headers, "FILEDATE", index=0, cast=int)
            month = first_header_value(headers, "FILEDATE", index=1, cast=int)
            day = first_header_value(headers, "FILEDATE", index=2, cast=int)
        else:
            return None
    else:
        g = re.search(r"#FILEDATE[\s=]*(\d+)[,\s+]+(\d+)[,\s+]+(\d+)", headers)
        if g:
            year = int(g.group(1))
            month = int(g.group(2))
            day = int(g.group(3))
        else:
            return None

    return date(year, month, day)


def parse_columns_number(headers):
    """
    Function that returns the columns number as an int.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: Columns number value.
    """
    if isinstance(headers, dict):
        return max([int(values[0]) for values in headers["COLUMNINFO"]])
    else:
        g = re.findall(r"#COLUMNINFO[=\s+]+(\d*)", headers)
        if g:
            return max(map(int, re.findall(r"#COLUMNINFO[=\s+]+(\d*)", headers)))


def parse_quantity_number(headers, column_number):
    """
    Function to parse the quantity number.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :param column_number: (int) Number of the column.
    :return: Quantity number.
    """
    # Convert the variable number to a string so it's cheaper to compare
    column_number_str = str(column_number)

    try:
        if isinstance(headers, dict):
            if "COLUMNINFO" in headers:
                # Loop over all headers to find the right number
                for values in headers["COLUMNINFO"]:
                    if values[0] == column_number_str:
                        return float(values[3])
        else:
            # Find all '#COLUMNINFO= **,' strings first
            for match in re.finditer(
                r"#COLUMNINFO[=\s+]+(\d+)+[,\s+][^,]*[,\s+]+[^,]*[,\s+]+(\d+)", headers
            ):
                # The first group is the column number
                if match.group(1) == column_number_str:
                    # The second group is the actual value
                    return cast_string(float, match.group(2))

    except ValueError:
        pass
    return None


def parse_column_info(headers, column_number, dictionary):
    """
    Function that returns the column info assigned to a quantity number.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :param column_number: (int) Number of the column.
    :param dictionary: (dict) Dictionary in which the quantity number is searched as a key.
    :return: Column info (value) of each quantity number (key).
    """
    try:
        quantity_number = parse_quantity_number(headers, column_number)
        column_info = dictionary[quantity_number]
    except KeyError:
        column_info = "column_code=" + str(column_number)

    return column_info


def parse_column_separator(headers):
    """
    Function to parse the column separator. It is used only in the borehole class.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: Column separator.
    """
    if isinstance(headers, dict):
        return first_header_value(headers, "COLUMNSEPARATOR")
    else:
        return parse_regex_cast(r"#COLUMNSEPARATOR+[=\s+]+(.)", headers, str, 1)


def parse_test_id(headers):
    """
    Function to parse the test id.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: test id.
    """
    result = None
    if isinstance(headers, dict):
        result = first_header_value(headers, "TESTID")
    else:
        result = parse_regex_cast(r"#TESTID+[=\s+]+(.*)", headers, str, 1)

    if result is not None:
        return result.strip()


def parse_record_separator(headers):
    """
    Function to parse the record separator(end of the line). It is used only in the borehole class.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: Record separator.
    """
    if isinstance(headers, dict):
        return first_header_value(headers, "RECORDSEPARATOR")
    else:
        return parse_regex_cast(r"#RECORDSEPARATOR+[=\s+]+(.)", headers, str, 1)


def find_separator(headers):
    """

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return:
    """
    # TODO: verify that this space is good enough for all use cases
    return parse_column_separator(headers) or " "


def parse_soil_code(s):
    """
    Function to parse the soil code.

    :param s: (str) String with the soil code.
    :return: Soil code.
    """
    return s.replace("'", "")


def create_soil_type(s):
    """
    Function to create the description of the soil type.

    :param s: (str) Soil code.
    :return: (str) Description of the soil code.
    """
    string_noquote = s.replace("'", "")
    split_letters = list(string_noquote)

    # split_soil_string = list(soil_string)
    dict_name = {"G": "gravel", "K": "clay", "L": "loam", "V": "peat", "Z": "sand"}
    dict_addition = {
        "z": "sand",
        "s": "silt",
        "m": "mineral",
        "k": "clay",
        "g": "gravel",
        "h": "humeus",
    }
    dict_intensity = {"1": 5, "2": 10, "3": 15, "4": 20}
    dict_exceptions = {
        "GM": "layer information missing",
        "NBE": "soil cannot be classified properly",
        "W": "water",
    }
    soil_name = ""
    try:
        if string_noquote != "":
            if string_noquote in dict_exceptions:
                soil_name = soil_name + dict_exceptions[string_noquote]
            else:
                sum_addition = 0
                percentage_main_component = 100
                main_component = dict_name[split_letters[0]]
                soil_additions = ""
                for i in range(1, len(split_letters)):
                    if split_letters[i] in dict_addition:
                        soil_additions = (
                            soil_additions + " with " + dict_addition[split_letters[i]]
                        )
                    elif split_letters[i] in dict_intensity:
                        soil_additions = (
                            soil_additions
                            + " "
                            + str(dict_intensity[split_letters[i]])
                            + "%"
                        )
                        sum_addition = sum_addition + dict_intensity[split_letters[i]]
                if sum_addition > 0:
                    percentage_main_component = percentage_main_component - sum_addition
                soil_name = (
                    main_component
                    + " "
                    + str(percentage_main_component)
                    + "%"
                    + soil_additions
                )
        else:
            soil_name = "soil_not_defined"
    except KeyError:
        soil_name = "soil_not_according_with_NEN_classification"
    return soil_name


INDICES = {"g": 0, "z": 1, "k": 2, "s": 5, "m": 1, "h": 4, "l": 3, "v": 4, "p": 4}

INTENSITY = {"1": 0.05, "2": 0.10, "3": 0.15, "4": 0.20}

NO_CLASSIFY = {
    "gm": "layer information missing",
    "nbe": "soil cannot be classified properly",
    "w": "water",
}


def soil_quantification(s):
    """
    Function to create the quantification of the soil type.

    :param s:(str) Soil code.
    :return: Quantification of the soil code.
    """
    #   0    1    2    3    4     5
    # ['G', 'S', 'C', 'L', 'P', 'SI']
    dist = np.zeros(6)

    s = s.replace("'", "").split(" ")[0].lower()

    if s in NO_CLASSIFY or len(s) == 0:
        return np.ones(6) * -1

    tokens = list(enumerate(s))
    numerics = dict(filter(lambda t: t[1].isnumeric(), tokens))
    alphabetics = dict(
        filter(lambda t: not t[1].isnumeric(), tokens[1:])
    )  # skip the first one.

    for i, token in alphabetics.items():
        idx = INDICES[token]
        if (i + 1) in numerics:
            v = INTENSITY[numerics[i + 1]]
        else:
            v = 0.05
        dist[idx] = v

    # Sometimes the same soil type has multiple tokens ?? e.g. Kkgh2.
    dist[INDICES[s[0]]] = 0
    dist[INDICES[s[0]]] = 1 - dist.sum()
    return dist


def parse_add_info(headers):
    """
    Function to parse all the additional informations.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: (str) Additional informations.
    """
    dict_add_info = {
        "DO": "dark ",
        "LI": "light ",
        "TBL": "blue-",
        "TBR": "brown-",
        "TGE": "yellow-",
        "TGN": "green-",
        "TGR": "gray-",
        "TOL": "olive-",
        "TOR": "orange-",
        "TPA": "violet-",
        "TRO": "red-",
        "TWI": "white-",
        "TRZ": "pink-",
        "TZW": "black-",
        "BL": "blue ",
        "BR": "brown ",
        "GE": "yellow ",
        "GN": "green ",
        "GR": "gray ",
        "OL": "olive ",
        "OR": "orange ",
        "PA": "violet ",
        "RO": "red ",
        "WI": "white ",
        "RZ": "pink ",
        "ZW": "black ",
        "ZUF": "uiterst fijn ",
        "ZZF": "zeer fijn ",
        "ZMF": "matig fijn ",
        "ZMG": "matig grof ",
        "ZZG": "zeer grof ",
        "ZUG": "uiterst grof ",
        "SZK": "zeer kleine spreiding ",
        "SMK": "matig kleine spreiding|",
        "SMG": "matig grote spreiding ",
        "SZG": "zeer grote spreiding ",
        "STW": "tweetoppige spreiding ",
        "ZZH": "sterk hoekig ",
        "ZHK": "hoekig ",
        "ZMH": "matig hoekig ",
        "ZMA": "afgerond ",
        "ZSA": "sterk afgerond ",
        "GFN": "fijn grind ",
        "GMG": "matig grof grind ",
        "GZG": "zeer grof grind ",
        "FN1": "spoor fijn grind (<1%) ",
        "FN2": "weinig fijn grind (1-25%) ",
        "FN3": "veel fijn grind (25-50%) ",
        "FN4": "zeer veel fijn grind (50-75%) ",
        "FN5": "uiterst veel fijn grind (>75%) ",
        "MG1": "spoor matig grof grind (<1%) ",
        "MG2": "weinig matig grof grind (1-25%) ",
        "MG3": "veel matig grof grind (25-50%) ",
        "MG4": "zeer veel matig grof grind (50-75%) ",
        "MG5": "uiterst veel matig grof grind(>75%) ",
        "GG1": "spoor zeer grof grind (<1%) ",
        "GG2": "weinig zeer grof grind (1-25%) ",
        "GG3": "veel zeer grof grind (25-50%) ",
        "GG4": "zeer veel zeer grof grind (50-75%) ",
        "GG5": "uiterst veel zeer grof grind (>75%) ",
        "AV1": "zwak amorf ",
        "AV2": "matig amorf ",
        "AV3": "sterk amorf ",
        "BSV": "bosveen ",
        "HEV": "heideveen ",
        "MOV": "mosveen ",
        "RIV": "rietveen ",
        "SZV": "Scheuchzeriaveen ",
        "VMV": "veenmosveen ",
        "WOV": "wollegrasveen ",
        "ZEV": "zeggeveen ",
        "KZSL": "klei zeer slap ",
        "KSLA": "klei slap ",
        "KMSL": "klei matig slap ",
        "KMST": "klei stevig ",
        "KZST": "klei zeer stevig ",
        "KHRD": "klei hard ",
        "KZHR": "klei zeer hard ",
        "LZSL": "leem zeer slap ",
        "LSLA": "leem slap ",
        "LMSL": "leem matig slap ",
        "LMST": "leem matig stevig ",
        "LSTV": "leem stevig ",
        "LZST": "leem zeer stevig ",
        "LHRD": "leem hard ",
        "LZHR": "leem zeer hard ",
        "VZSL": "veen zeer slap ",
        "VSLA": "veen slap ",
        "VMSL": "veen matig slap ",
        "VMST": "veen matig stevig ",
        "VSTV": "stevig ",
        "LOS": "los gepakt ",
        "NOR": "normaal gepakt ",
        "VAS": "vast gepakt ",
        "VGZZ": "zeer zacht ",
        "VGZA": "zacht ",
        "VGMZ": "matig zacht ",
        "VGMH": "matig hard ",
        "VGHA": "hard ",
        "VGZH": "zeer hard ",
        "VGEH": "extreem hard ",
        "SCH0": "geen schelpmateriaal 0% ",
        "SCH1": "spoor schelpmateriaal <1% ",
        "SCH2": "weinig schelpmateriaal (1-10%) ",
        "SCH3": "veel schelpmateriaal (10-30%) ",
        "CA1": "kalkloos ",
        "CA2": "kalkarm ",
        "CA3": "kalkrijk ",
        "GC0": "geen glauconiet 0% ",
        "GC1": "spoor glauconiet <1% ",
        "GC2": "weinig glauconiet (1-10%) ",
        "GC3": "veel glauconiet (10-30%) ",
        "GC4": "zeer veel glauconiet(30-50%) ",
        "GC5": "uiterst veel glauconiet (>50%) ",
        "BST1": "spoor baksteen ",
        "BST2": "weinig baksteen ",
        "BST3": "veel baksteen ",
        "PUR1": "spoor puinresten ",
        "PUR2": "weinig puinresten ",
        "PUR3": "veel puinresten ",
        "SIN1": "spoor sintels ",
        "SIN2": "weinig sintels ",
        "SIN3": "veel sintels ",
        "STO1": "spoor stortsteen ",
        "STO2": "weinig stortsteen ",
        "STO3": "veel stortsteen ",
        "VUI1": "spoor vuilnis ",
        "VUI2": "weinig vuilnis ",
        "VUI3": "veel vuilnis ",
        "AF": "afval ",
        "AS": "asfalt ",
        "BE": "beton ",
        "BI": "bitumen ",
        "BT": "ballast ",
        "BST": "baksteen ",
        "GI": "gips ",
        "GA": "glas ",
        "HK": "houtskool ",
        "HU": "huisvuil ",
        "KA": "kalk ",
        "KG": "kolengruis ",
        "KO": "kolen ",
        "KT": "krijt ",
        "ME": "metaal ",
        "MI": "mijnsteen ",
        "OE": "oer ",
        "PL": "planten ",
        "PU": "puin ",
        "SI": "sintels ",
        "SL": "slakken ",
        "WO": "wortels ",
        "YZ": "ijzer ",
        "GL": "gley ",
        "RT": "roest ",
        "SE": "silex ",
        "BIO": "bioturbatie ",
        "DWO": "doorworteling ",
        "GCM": "cm-gelaagdheid ",
        "GDM": "dm-gelaagdheid ",
        "GDU": "dubbeltjes-gelaagdheid ",
        "GMM": "mm-gelaagdheid ",
        "GRG": "graafgangen ",
        "GSC": "scheve gelaagdheid ",
        "GSP": "spekkoek-gelaagdheid ",
        "HOM": "homogeen ",
        "GE1": "zwak gelaagd ",
        "GE2": "weinig gelaagd ",
        "GE3": "sterk gelaagd ",
        "GEX": "gelaagd ",
        "STGL": "met grindlagen ",
        "STKL": "met kleilagen ",
        "STLL": "met leemlagen ",
        "STSL": "met stenenlagen ",
        "STVL": "met veenlagen ",
        "STZL": "met zandlagen ",
        "STBR": "met bruinkoollagen ",
        "STDE": "met detrituslagen ",
        "STGY": "met gyttjalagen ",
        "STSC": "met schelpenlagen ",
        "ANT": "Antropogeen ",
        "BOO": "Boomse klei ",
        "DEZ": "dekzand ",
        "KEL": "keileem ",
        "LSS": "loess ",
        "POK": "potklei ",
        "WAR": "warven ",
        "DR": "Formatie van Drente ",
        "EC": "Formatie van Echteld ",
        "KR": "Formatie van Kreftenheye ",
        "NA": "Formatie van Naaldwijk ",
        "NI": "Formatie van Nieuwkoop ",
        "TW": "Formatie van Twente ",
        "WA": "Formatie van Waalre ",
    }

    if isinstance(headers, dict):
        raise Exception("todo")
    else:
        string_noquote = headers[1:-1]
        string2list = string_noquote.split("''")
        add_info = ""
        for i, string in enumerate(string2list):
            if string:
                add_info += "{}) ".format(i + 1) + "".join(
                    [
                        dict_add_info[string_split]
                        if string_split in dict_add_info
                        else string_split + " "
                        for string_split in string.split(" ")
                    ]
                )
        return add_info


def assign_multiple_columns(df, columns, partial_df):
    return df.drop(columns).hstack(partial_df[columns])


def kpa_to_mpa(df, columns):
    return assign_multiple_columns(df, columns, df[columns] * 10 ** -3)


def none_to_zero(df):
    return df.fill_null(0.0)


def nap_to_depth(zid, nap):
    return -(nap - zid)


def depth_to_nap(depth, zid):
    return zid - depth


def join_gef(bore, cpt):
    """
    Join a cpt and bore file in one Dataframe based on depth.

    :param bore: (ParseBORE)
    :param cpt: (ParseCPT)
    :return: (pl.DataFrame)
    """
    df_cpt = cpt.df.assign(join_idx=0)
    df_bore = bore.df.loc[
        bore.df[["G", "S", "C", "L", "P", "SI"]].sum(1) == 1
    ].reset_index(drop=True)

    df_cpt = df_cpt[df_cpt["depth"] > df_bore["depth_top"].min()].reset_index(drop=True)
    idx = np.searchsorted(df_cpt["depth"].values, df_bore["depth_top"].values)

    a = np.zeros(df_cpt.shape[0])
    for i in range(len(idx) - 1):
        a[idx[i] : idx[i + 1]] = i

    a[idx[i + 1] :] = i + 1
    df_cpt["join_idx"] = a

    return df_cpt.merge(
        df_bore[["soil_code", "G", "S", "C", "L", "P", "SI"]].reset_index(-1),
        left_on="join_idx",
        right_on="index",
    ).drop(["index", "join_idx"], axis=1)



import io
import logging
import re
from typing import List

import numpy as np
import polars as pl
import pygef.utils as utils
from polars import col, lit, when

# Try to import the optimized Rust header parsing but if that doesn't succeed
# use the built-in python regex methods
try:
    import gef

    USE_RUST_PARSED_HEADERS = False
except ImportError:
    USE_RUST_PARSED_HEADERS = False

logger = logging.getLogger(__name__)


MAP_QUANTITY_NUMBER_COLUMN_NAME_CPT = {
    1: "penetration_length",
    2: "qc",  # 2
    3: "fs",  # 3
    4: "friction_number",  # 4
    5: "u1",  # 5
    6: "u2",  # 6
    7: "u3",  # 7
    8: "inclination",  # 8
    9: "inclination_ns",  # 9
    10: "inclination_ew",  # 10
    11: "corrected_depth",  # 11
    12: "time",  # 12
    13: "corrected_qc",  # 13
    14: "net_cone_resistance",  # 14
    15: "pore_ratio",  # 15
    16: "cone_resistance_number",  # 16
    17: "weight_per_unit_volume",  # 17
    18: "initial_pore_pressure",  # 18
    19: "total_vertical_soil_pressure",  # 19
    20: "effective_vertical_soil_pressure",
    21: "inclination_in_x_direction",
    22: "inclination_in_y_direction",
    23: "electric_conductivity",
    31: "magnetic_field_x",
    32: "magnetic_field_y",
    33: "magnetic_field_z",
    34: "total_magnetic_field",
    35: "magnetic_inclination",
    36: "magnetic_declination",
    99: "classification_zone_robertson_1990",
    # found in:#COMPANYID= Fugro GeoServices B.V., NL005621409B08, 31
    131: "speed",  # found in:COMPANYID= Multiconsult, 09073590, 31
    135: "Temperature_c",  # found in:#COMPANYID= Inpijn-Blokpoel,
    250: "magneto_slope_y",  # found in:COMPANYID= Danny, Tjaden, 31
    251: "magneto_slope_x",
}  # found in:COMPANYID= Danny, Tjaden, 31

COLUMN_NAMES_BORE = [
    "depth_top",  # 1
    "depth_bottom",  # 2
    "lutum_percentage",  # 3
    "silt_percentage",  # 4
    "sand_percentage",  # 5
    "gravel_percentage",  # 6
    "organic_matter_percentage",  # 7
    "sand_median",  # 8
    "gravel_median",
]  # 9
MAP_QUANTITY_NUMBER_COLUMN_NAME_BORE = dict(enumerate(COLUMN_NAMES_BORE, 1))

COLUMN_NAMES_BORE_CHILD = [
    "depth_top",  # 1
    "depth_bottom",  # 2
    "undrained_shear_strength",  # 3
    "vertical_permeability",  # 4
    "horizontal_permeability",  # 5
    "effective_cohesion_at_x%_strain",  # 6
    "friction_angle_at_x%_strain",  # 7
    "water_content",  # 8
    "dry_volumetric_mass",  # 9
    "wet_volumetric_mass",  # 10
    "d_50",  # 11
    "d_60/d_10_uniformity",  # 12
    "d_90/d_10_gradation",  # 13
    "dry_volumetric_weight",  # 14
    "wet_volumetric_weight",  # 15
    "vertical_strain",
]  # 16
MAP_QUANTITY_NUMBER_COLUMN_NAME_BORE_CHILD = dict(enumerate(COLUMN_NAMES_BORE_CHILD, 1))

dict_soil_type_rob = {
    "Peat": 1,
    "Clays - silty clay to clay": 2,
    "Silt mixtures - clayey silt to silty clay": 3,
    "Sand mixtures - silty sand to sandy silt": 4,
    "Sands - clean sand to silty sand": 5,
    "Gravelly sand to dense sand": 6,
}

dict_soil_type_been = {
    "Peat": 1,
    "Clays": 2,
    "Clayey silt to silty clay": 3,
    "Silty sand to sandy silt": 4,
    "Sands: clean sand to silty": 5,
    "Gravelly sands": 6,
}


class _Gef:
    """
    The gef parser is built following the conventional format described in:
    https://publicwiki.deltares.nl/download/attachments/102204318/GEF-CPT.pdf?version=1&modificationDate=1409732008000&api=v2
    """

    def __init__(self, path=None, string=None):
        """
        Base class of gef parser. It switches between the cpt or borehole parser.

        It takes as input either the path to the gef file or the gef file as string.

        Parameters
        ----------
        path: str
            Path to the *.gef file.
        string: str
            String version of the *.gef file.

        """
        self.path = path
        self.df = None
        self.net_surface_area_quotient_of_the_cone_tip = None
        self.pre_excavated_depth = None

        if string is None:
            with open(path, encoding="utf-8", errors="ignore") as f:
                string = f.read()

        self.s = string

        if USE_RUST_PARSED_HEADERS:
            # Use the Rust optimized header parser
            self._data, self._headers = gef.parse(string)
        else:
            # Use the fallback python regex parser
            end_of_header = utils.parse_end_of_header(string)
            self._headers, self._data = string.split(end_of_header)

        self.zid = utils.parse_zid_as_float(self._headers)
        self.height_system = utils.parse_height_system(self._headers)
        self.x = utils.parse_xid_as_float(self._headers)
        self.y = utils.parse_yid_as_float(self._headers)
        self.file_date = utils.parse_file_date(self._headers)
        self.test_id = utils.parse_test_id(self._headers)
        self.type = utils.parse_gef_type(self._headers)


class _GefCpt(_Gef):
    def __init__(self, path=None, string=None):
        """
        Parser of the cpt file.

        Parameters
        ----------
        path: str
            Path to the *.gef file.
        string: str
            String version of the *.gef file.
        """
        super().__init__(path=path, string=string)
        if not self.type == "cpt":
            raise ValueError(
                "The selected gef file is not a cpt. "
                "Check the REPORTCODE or the PROCEDURECODE."
            )
        self.project_id = utils.parse_project_type(self._headers, "cpt")
        self.cone_id = utils.parse_cone_id(self._headers)
        self.cpt_class = utils.parse_cpt_class(self._headers)
        self.column_void = utils.parse_column_void(self._headers)
        self.nom_surface_area_cone_tip = utils.parse_measurement_var_as_float(
            self._headers, 1
        )
        self.nom_surface_area_friction_element = utils.parse_measurement_var_as_float(
            self._headers, 2
        )
        self.net_surface_area_quotient_of_the_cone_tip = (
            utils.parse_measurement_var_as_float(self._headers, 3)
        )
        self.net_surface_area_quotient_of_the_friction_casing = (
            utils.parse_measurement_var_as_float(self._headers, 4)
        )
        self.distance_between_cone_and_centre_of_friction_casing = (
            utils.parse_measurement_var_as_float(self._headers, 5)
        )
        self.friction_present = utils.parse_measurement_var_as_float(self._headers, 6)
        self.ppt_u1_present = utils.parse_measurement_var_as_float(self._headers, 7)
        self.ppt_u2_present = utils.parse_measurement_var_as_float(self._headers, 8)
        self.ppt_u3_present = utils.parse_measurement_var_as_float(self._headers, 9)
        self.inclination_measurement_present = utils.parse_measurement_var_as_float(
            self._headers, 10
        )
        self.use_of_back_flow_compensator = utils.parse_measurement_var_as_float(
            self._headers, 11
        )
        self.type_of_cone_penetration_test = utils.parse_measurement_var_as_float(
            self._headers, 12
        )
        self.pre_excavated_depth = utils.parse_measurement_var_as_float(
            self._headers, 13
        )
        self.groundwater_level = utils.parse_measurement_var_as_float(self._headers, 14)
        self.water_depth_offshore_activities = utils.parse_measurement_var_as_float(
            self._headers, 15
        )
        self.end_depth_of_penetration_test = utils.parse_measurement_var_as_float(
            self._headers, 16
        )
        self.stop_criteria = utils.parse_measurement_var_as_float(self._headers, 17)
        self.zero_measurement_cone_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 20)
        )
        self.zero_measurement_cone_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 21)
        )
        self.zero_measurement_friction_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 22)
        )
        self.zero_measurement_friction_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 23)
        )
        self.zero_measurement_ppt_u1_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 24)
        )
        self.zero_measurement_ppt_u1_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 25)
        )
        self.zero_measurement_ppt_u2_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 26)
        )
        self.zero_measurement_ppt_u2_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 27)
        )
        self.zero_measurement_ppt_u3_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 28)
        )
        self.zero_measurement_ppt_u3_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 29)
        )
        self.zero_measurement_inclination_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 30)
        )
        self.zero_measurement_inclination_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 31)
        )
        self.zero_measurement_inclination_ns_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 32)
        )
        self.zero_measurement_inclination_ns_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 33)
        )
        self.zero_measurement_inclination_ew_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 34)
        )
        self.zero_measurement_inclination_ew_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 35)
        )
        self.mileage = utils.parse_measurement_var_as_float(self._headers, 41)

        column_names = determine_column_names(self._headers)

        self.df = (
            self.parse_data(self._headers, self._data, column_names)
            .lazy()
            .pipe(replace_column_void, self.column_void)
            .pipe(correct_pre_excavated_depth, self.pre_excavated_depth)
            .with_column(correct_depth_with_inclination(column_names))
            .select(
                # Remove None values since they throw an error
                [
                    expr
                    for expr in [
                        pl.all().exclude(["depth", "friction_number"]),
                        col("depth").abs(),
                        calculate_friction_number(column_names),
                        self.calculate_elevation_with_respect_to_nap(
                            self.zid, self.height_system
                        ),
                    ]
                    if expr is not None
                ]
            )
            .collect()
        )

    @staticmethod
    def calculate_elevation_with_respect_to_nap(zid, height_system):
        if zid is not None and height_system == 31000.0:
            return (zid - pl.col("depth")).alias("elevation_with_respect_to_nap")

        return None

    @staticmethod
    def parse_data(headers, data_s, column_names=None):
        separator = utils.find_separator(headers)

        # Remove multiple whitespaces
        # TODO: find a way for polars to handle columns with variable amounts of whitespace
        if separator == " ":
            new_data = re.sub("[ \t]+", " ", data_s.replace("!", ""))
        else:
            # If we have another separator remove all whitespace around it
            new_data = re.sub(
                f"[\t ]*{re.escape(separator)}[\t ]*",
                separator,
                data_s.replace(separator + "!", "").replace("!", ""),
            )

        # Remove whitespace at the beginning and end of lines, and remove the
        # last trailing line
        new_data = "\n".join([line.strip() for line in new_data.splitlines()]).rstrip()

        return pl.read_csv(
            new_data.encode(),
            sep=separator,
            new_columns=column_names,
            has_headers=False,
        )


class _GefBore(_Gef):
    def __init__(self, path=None, string=None):
        """
        Parser of the borehole file.

        Parameters
        ----------
        path: str
            Path to the *.gef file.
        string: str
            String version of the *.gef file.
        """
        super().__init__(path=path, string=string)
        if self.type == "bore":
            pass
        elif self.type == "borehole-report":
            raise ValueError(
                "The selected gef file is a GEF-BOREHOLE-Report. Can only parse "
                "GEF-CPT-Report and GEF-BORE-Report. Check the PROCEDURECODE."
            )
        else:
            raise ValueError(
                "The selected gef file is not a borehole. "
                "Check the REPORTCODE or the PROCEDURECODE."
            )

        self.project_id = utils.parse_project_type(self._headers, "bore")
        self.nen_version = "NEN 5104"

        # This is usually not correct for the boringen
        columns_number = utils.parse_columns_number(self._headers)
        column_separator = utils.parse_column_separator(self._headers)
        record_separator = utils.parse_record_separator(self._headers)
        data_s_rows = self._data.split(record_separator)
        data_rows_soil = self.extract_soil_info(
            data_s_rows, columns_number, column_separator
        )

        self.df = (
            self.parse_data_column_info(
                self._headers, self._data, column_separator, columns_number
            )
            .pipe(self.parse_data_soil_code, data_rows_soil)
            .pipe(self.parse_data_soil_type, data_rows_soil)
            .pipe(self.parse_add_info_as_string, data_rows_soil)
            .pipe(self.parse_soil_quantification, data_rows_soil)
        )

        # Drop the columns if they exist, do nothing if they don't
        for column in [
            "sand_median",
            "gravel_median",
            "lutum_percentage",
            "silt_percentage",
            "sand_percentage",
            "gravel_percentage",
            "organic_matter_percentage",
            "soil_type",
        ]:
            if column in self.df.columns:
                self.df.drop_in_place(column)

    @staticmethod
    def parse_add_info_as_string(df, data_rows_soil):
        df["remarks"] = [
            utils.parse_add_info("".join(row[1::])) for row in data_rows_soil
        ]

        return df

    @staticmethod
    def extract_soil_info(data_s_rows, columns_number, column_separator):
        return list(
            map(
                lambda x: x.split(column_separator)[columns_number:-1], data_s_rows[:-1]
            )
        )

    @staticmethod
    def parse_data_column_info(headers, data_s, sep, columns_number, columns_info=None):
        if columns_info is None:
            col = list(
                map(
                    lambda x: utils.parse_column_info(
                        headers, x, MAP_QUANTITY_NUMBER_COLUMN_NAME_BORE
                    ),
                    range(1, columns_number + 1),
                )
            )
            return pl.read_csv(
                io.StringIO(data_s),
                sep=sep,
                new_columns=col,
                has_headers=False,
                projection=list(range(0, len(col))),
            )
        else:
            return pl.read_csv(
                io.StringIO(data_s),
                sep=sep,
                new_columns=columns_info,
                has_headers=False,
            )

    @staticmethod
    def parse_data_soil_type(df, data_rows_soil):
        df["soil_type"] = list(
            map(lambda x: utils.create_soil_type(x[0]), data_rows_soil)
        )

        return df

    @staticmethod
    def parse_data_soil_code(df, data_rows_soil):
        df["soil_code"] = list(
            map(lambda x: utils.parse_soil_code(x[0]), data_rows_soil)
        )

        return df

    @staticmethod
    def parse_soil_quantification(df, data_rows_soil):
        data = np.array([utils.soil_quantification(x[0]) for x in data_rows_soil])

        # Gravel
        df["gravel_component"] = data[:, 0]
        # Sand
        df["sand_component"] = data[:, 1]
        # Clay
        df["clay_component"] = data[:, 2]
        # Loam
        df["loam_component"] = data[:, 3]
        # Peat
        df["peat_component"] = data[:, 4]
        # Silt
        df["silt_component"] = data[:, 5]

        return df


def replace_column_void(lf: pl.LazyFrame, column_void) -> pl.LazyFrame:
    if column_void is None:
        return lf

    # TODO: what to do with multiple columnvoids?
    if isinstance(column_void, list):
        column_void = column_void[0]

    return (
        # Get all values matching column_void and change them to null
        lf.select(
            pl.when(pl.all() == pl.lit(column_void))
            .then(pl.lit(None))
            .otherwise(pl.all())
            .keep_name()
        )
        # Interpolate all null values
        .select(pl.all().interpolate())
        # Remove the rows with null values
        .drop_nulls()
    )


def correct_pre_excavated_depth(lf: pl.LazyFrame, pre_excavated_depth) -> pl.LazyFrame:
    if pre_excavated_depth is not None and pre_excavated_depth > 0:
        return lf.filter(col("penetration_length") >= pre_excavated_depth)
    return lf


def correct_depth_with_inclination(columns):
    """
    Return the expression needed to correct depth
    """
    if "corrected_depth" in columns:
        return col("corrected_depth").alias("depth")
    elif "inclination" in columns:
        pt = "penetration_length"

        # every different in depth needs to be corrected with the angle
        correction_factor = np.cos(
            np.radians(col("inclination").cast(pl.Float32).fill_null(0))
        )

        corrected_depth = (correction_factor * col(pt).diff()).cumsum()
        return (
            pl.when(corrected_depth.is_null())
            .then(col(pt))
            .otherwise(corrected_depth)
            .alias("depth")
        )
    else:
        return col("penetration_length").alias("depth")


def determine_column_names(headers, columns_number=None, columns_info=None):
    if columns_number is None and columns_info is None:
        columns_number = utils.parse_columns_number(headers)
        if columns_number is not None:
            columns_info = []
            for column_number in range(1, columns_number + 1):
                columns_info.append(
                    utils.parse_column_info(
                        headers, column_number, MAP_QUANTITY_NUMBER_COLUMN_NAME_CPT
                    )
                )

    return columns_info


def calculate_friction_number(column_names: List[str]) -> "pl.Expr":
    if "fs" in column_names and "qc" in column_names:
        return (
            col("fs") / when(col("qc") == 0.0).then(None).otherwise(col("qc")) * 100.0
        ).alias("friction_number")
    else:
        return lit(0.0).alias("friction_number")



import numpy as np
import polars as pl
import pygef.utils as utils
from pygef import geo


def n_exponent(df, p_a):
    df["n"] = (
        0.381 * df["type_index_n"] + 0.05 * (df["effective_soil_pressure"] / p_a) - 0.15
    )

    # Set the maximum for N to 1.0
    df[df["n"] > 1.0, "n"] = 1.0

    return df


def normalized_cone_resistance_n(df, p_a):
    # We have to convert the series to numpy because they don't support pow
    df["normalized_cone_resistance"] = (
        (df["qt"] - df["soil_pressure"])
        / p_a
        * (p_a / df["effective_soil_pressure"]) ** df["n"]
    )
    df[df["normalized_cone_resistance"] < 0.0, "normalized_cone_resistance"] = 1.0

    return df


def type_index(df):
    # We have to convert the series to numpy because they don't support pow
    df["type_index"] = (
        (3.47 - np.log10(df["normalized_cone_resistance"])) ** 2.0
        + (np.log10(df["normalized_friction_ratio"]) + 1.22) ** 2.0
    ) ** 0.5

    return df


def ic_to_gamma(df, water_level):
    """
    Assign the right gamma (unit soil weight kN/m^3) to the corresponding Ic.

    :param df: (DataFrame) Original DataFrame.
    :param water_level: (int) Water level with respect to ground level.
    :return: Updated DataFrame.
    """
    mask_below_water = (1.0 - df["depth"]) < water_level
    # TODO: how to fill it properly with the same initial values?
    df["gamma_predict"] = np.tile(1.0, len(df.rows()))

    ic_mask = df["type_index"] > 3.6
    # gamma_(sat) and ic > 3.6
    df[ic_mask, "gamma_predict"] = 11.0

    ic_mask = df["type_index"] <= 3.6
    # gamma_(sat) and ic < 3.6
    df[ic_mask, "gamma_predict"] = 16.0

    ic_mask = df["type_index"] <= 2.95
    # gamma_(sat) and ic < x
    df[ic_mask, "gamma_predict"] = 18.0

    ic_mask = df["type_index"] <= 2.6
    # gamma_sat and ic < x
    df[ic_mask & mask_below_water, "gamma_predict"] = 19.0

    ic_mask = df["type_index"] <= 2.05
    # gamma_sat and ic < x
    df[ic_mask & mask_below_water, "gamma_predict"] = 20.0

    return df


def ic_to_soil_type(df):
    """
    Assign the soil type to the corresponding Ic.

    :param df: (DataFrame) Original DataFrame.
    :return: (DataFrame) Updated DataFrame.
    """
    # TODO: how to fill it properly with the same initial values?
    df["soil_type"] = np.tile("", len(df.rows()))

    ic_mask = df["type_index"] > 3.6
    df[ic_mask, "soil_type"] = "Peat"

    ic_mask = df["type_index"] <= 3.6
    df[ic_mask, "soil_type"] = "Clays - silty clay to clay"

    ic_mask = df["type_index"] <= 2.95
    df[ic_mask, "soil_type"] = "Silt mixtures - clayey silt to silty clay"

    ic_mask = df["type_index"] <= 2.6
    df[ic_mask, "soil_type"] = "Sand mixtures - silty sand to sandy silt"

    ic_mask = df["type_index"] <= 2.05
    df[ic_mask, "soil_type"] = "Sands - clean sand to silty sand"

    ic_mask = df["type_index"] <= 1.31
    df[ic_mask, "soil_type"] = "Gravelly sand to dense sand"

    return df


def none_to_zero(df):
    return df.fill_null(0)


def iterate_robertson(
    original_df,
    water_level,
    new=True,
    area_quotient_cone_tip=None,
    pre_excavated_depth=None,
    p_a=0.1,
):
    """
    Iteration function for Robertson classifier.

    :param original_df: (DataFrame)
    :param water_level: (int) Water level with respect to ground level.
    :param new: (bool) True to use the new classification, False otherwise. Default: True
    :param area_quotient_cone_tip: (float)
    :param pre_excavated_depth: (float)
    :param p_a: (float) Atmospheric pressure. Default: 0.1 MPa.
    :return: (DataFrame)
    """
    gamma = np.ones(original_df.shape[0]) * 18.0
    n = np.ones(original_df.shape[0])
    type_index_n = np.ones(original_df.shape[0])

    c = 0
    while True:
        c += 1
        if new:
            df = original_df
            df["n"] = n
            df["type_index_n"] = type_index_n
            df["gamma"] = gamma
            f = new_robertson

            def condition(x):
                return np.all(
                    x["gamma_predict"].series_equal(pl.Series(gamma))
                    and np.all(x["n"].series_equal(pl.Series(n)))
                )

        else:
            df = original_df
            df["gamma"] = gamma
            f = old_robertson

            def condition(x):
                return np.all(x["gamma_predict"].series_equal(pl.Series(gamma)))

        df = f(
            df,
            water_level,
            area_quotient_cone_tip=area_quotient_cone_tip,
            pre_excavated_depth=pre_excavated_depth,
            p_a=p_a,
        )

        df["gamma_predict"] = np.nan_to_num(df["gamma_predict"])
        if condition(df):
            break
        elif c == 4:
            break
        else:
            gamma = df["gamma_predict"]
            if new:
                n = df["n"]

    return df.pipe(ic_to_soil_type)


def old_robertson(
    df, water_level, area_quotient_cone_tip=None, pre_excavated_depth=None
):
    """
    Old (1990) implementation of Robertson.

    :param df: (DataFrame)
    :param water_level: (float)
    :param area_quotient_cone_tip: (float)
    :param pre_excavated_depth: (float)
    :return: (DataFrame) Classified dataframe.
    """
    df = (
        df.pipe(geo.delta_depth, pre_excavated_depth)
        .pipe(geo.soil_pressure)
        .pipe(geo.qt, area_quotient_cone_tip)
        .pipe(geo.water_pressure, water_level)
        .pipe(geo.effective_soil_pressure)
        .pipe(
            utils.kpa_to_mpa,
            ["soil_pressure", "effective_soil_pressure", "water_pressure"],
        )
        .pipe(geo.normalized_cone_resistance)
        .pipe(geo.normalized_friction_ratio)
        .pipe(utils.none_to_zero)
        .pipe(type_index)
        .pipe(ic_to_gamma, water_level)
    )

    return df


def new_robertson(
    df, water_level, area_quotient_cone_tip=None, pre_excavated_depth=None, p_a=0.1
):
    """
    New (2016) implementation of Robertson.

    :param df: (DataFrame)
    :param water_level: (float)
    :param area_quotient_cone_tip: (float)
    :param pre_excavated_depth: (float)
    :param p_a: (float)
    :return: (DataFrame) Classified dataframe.
    """
    df = (
        df.pipe(geo.delta_depth, pre_excavated_depth)
        .pipe(geo.soil_pressure)
        .pipe(geo.qt, area_quotient_cone_tip)
        .pipe(geo.water_pressure, water_level)
        .pipe(geo.effective_soil_pressure)
        .pipe(
            utils.kpa_to_mpa,
            ["soil_pressure", "effective_soil_pressure", "water_pressure"],
        )
        .pipe(n_exponent, p_a)
        .pipe(normalized_cone_resistance_n, p_a)
        .pipe(geo.normalized_friction_ratio)
        .pipe(utils.none_to_zero)
        .pipe(type_index)
        .pipe(ic_to_gamma, water_level)
    )
    return df



# pylint: disable=E1136
import os
import unittest
from datetime import datetime

import numpy as np
import pandas as pd
import polars as pl
import pygef.geo as geo
import pygef.utils as utils
from pygef.bore import Bore
from pygef.cpt import Cpt
from pygef.gef import (
    MAP_QUANTITY_NUMBER_COLUMN_NAME_CPT,
    _GefBore,
    _GefCpt,
    calculate_friction_number,
    correct_depth_with_inclination,
    correct_pre_excavated_depth,
    replace_column_void,
)


class GefTest(unittest.TestCase):
    def test_xy(self):
        cpt3 = Cpt("./pygef/test_files/cpt3.gef")
        self.assertEqual(cpt3.x, 110885)
        self.assertEqual(cpt3.y, 493345)

    def test_measurement_var_with_minus_sign(self):
        s = r"#MEASUREMENTVAR= 41, -15.000000, "
        v = utils.parse_measurement_var_as_float(s, 41)
        self.assertAlmostEqual(v, -15)

        h = {"MEASUREMENTVAR": [["41", "-15.000000"]]}
        v = utils.parse_measurement_var_as_float(h, 41)
        self.assertAlmostEqual(v, -15)

    def test_measurement_var_without_minus_sign(self):
        s = r"#MEASUREMENTVAR= 41, 15.000000, "
        v = utils.parse_measurement_var_as_float(s, 41)
        self.assertAlmostEqual(v, 15)

        h = {"MEASUREMENTVAR": [["41", "15.000000", ""]]}
        v = utils.parse_measurement_var_as_float(h, 41)
        self.assertAlmostEqual(v, 15)

    def test_measurement_var_integer(self):
        s = r"#MEASUREMENTVAR= 41, 0, deg, "
        v = utils.parse_measurement_var_as_float(s, 41)
        self.assertAlmostEqual(v, 0.0)

        h = {"MEASUREMENTVAR": [["41", "0", "deg", ""]]}
        v = utils.parse_measurement_var_as_float(h, 41)
        self.assertAlmostEqual(v, 0.0)

    def test_measurement_var_big_integer(self):
        s = r"#MEASUREMENTVAR= 41, 10000, deg, "
        v = utils.parse_measurement_var_as_float(s, 41)
        self.assertAlmostEqual(v, 10000)

        h = {"MEASUREMENTVAR": [["41", "10000", "deg", ""]]}
        v = utils.parse_measurement_var_as_float(h, 41)
        self.assertAlmostEqual(v, 10000)

    def test_measurement_var_different_space(self):
        s = r"#MEASUREMENTVAR = 41, 0, deg, "
        v = utils.parse_measurement_var_as_float(s, 41)
        self.assertAlmostEqual(v, 0.0)

        h = {"MEASUREMENTVAR": [["41", "0", "deg", ""]]}
        v = utils.parse_measurement_var_as_float(h, 41)
        self.assertAlmostEqual(v, 0.0)

    def test_measurement_var_different_comma(self):
        s = r"#MEASUREMENTVAR= 41 , 0, deg, "
        v = utils.parse_measurement_var_as_float(s, 41)
        self.assertAlmostEqual(v, 0.0)

        h = {"MEASUREMENTVAR": [["41", "0", "deg", ""]]}
        v = utils.parse_measurement_var_as_float(h, 41)
        self.assertAlmostEqual(v, 0.0)

    def test_parse_cpt_class(self):
        s = r"#MEASUREMENTTEXT= 6, NEN 5140 / klasse onbekend, sondeernorm en kwaliteitsklasse"
        v = utils.parse_cpt_class(s)
        self.assertEqual(v, None)

        h = {
            "MEASUREMENTTEXT": [
                ["6", "NEN 5140 / klasse onbekend", "sondeernorm en kwaliteitsklasse"]
            ]
        }
        v = utils.parse_cpt_class(h)
        self.assertEqual(v, None)

        s = r"#MEASUREMENTTEXT= 6, NEN-EN-ISO22476-1 / klasse 2 / TE2, gehanteerde norm en klasse en type sondering"
        v = utils.parse_cpt_class(s)
        self.assertEqual(v, 2.0)

        h = {
            "MEASUREMENTTEXT": [
                [
                    "6",
                    "NEN-EN-ISO22476-1 / klasse 2 / TE2",
                    "gehanteerde norm en klasse en type sondering",
                ]
            ]
        }
        v = utils.parse_cpt_class(h)
        self.assertEqual(v, 2.0)

        s = r"#MEASUREMENTTEXT= 6, Norm : NEN 5140; Klasse : 2, De norm waaraan deze sondering moet voldoen."
        v = utils.parse_cpt_class(s)
        self.assertEqual(v, 2.0)

        h = {
            "MEASUREMENTTEXT": [
                [
                    "6",
                    "Norm : NEN 5140; Klasse : 2",
                    "De norm waaraan deze sondering moet voldoen.",
                ]
            ]
        }
        v = utils.parse_cpt_class(h)
        self.assertEqual(v, 2.0)

        s = r"#MEASUREMENTTEXT= 6, Norm : NEN 5140; class : 2, De norm waaraan deze sondering moet voldoen."
        v = utils.parse_cpt_class(s)
        self.assertEqual(v, 2.0)

        h = {
            "MEASUREMENTTEXT": [
                [
                    "6",
                    "Norm : NEN 5140; class : 2",
                    "De norm waaraan deze sondering moet voldoen.",
                ]
            ]
        }
        v = utils.parse_cpt_class(s)
        self.assertEqual(v, 2.0)

    def test_file_date(self):
        s = r"#FILEDATE= 2004, 1, 14"
        v = utils.parse_file_date(s)
        self.assertTrue(v == datetime(2004, 1, 14).date())

        h = {"FILEDATE": [["2004", "1", "14"]]}
        v = utils.parse_file_date(h)
        self.assertTrue(v == datetime(2004, 1, 14).date())

    def test_project_id(self):
        s = r"#PROJECTID= CPT, 146203"
        v = utils.parse_project_type(s, "cpt")
        self.assertEqual(v, "146203")

        h = {"PROJECTID": [["CPT", "146203"]]}
        v = utils.parse_project_type(h, "cpt")
        self.assertEqual(v, "146203")

        s = r"#PROJECTID = DINO-BOR"
        v = utils.parse_project_type(s, "bore")
        self.assertEqual(v, "DINO-BOR")

        h = {"PROJECTID": [["DINO-BOR"]]}
        v = utils.parse_project_type(h, "bore")
        self.assertEqual(v, "DINO-BOR")

        s = r"#PROJECTID = CPT, 1018-0347-000, -"
        v = utils.parse_project_type(s, "cpt")
        self.assertEqual(v, "1018-0347-000")

        h = {"PROJECTID": [["CPT", "1018-0347-000", "-"]]}
        v = utils.parse_project_type(h, "cpt")
        self.assertEqual(v, "1018-0347-000")

    def test_zid(self):
        s = r"#ZID= 31000, 1.3, 0.0"
        v = utils.parse_zid_as_float(s)
        self.assertEqual(v, 1.3)

        h = {"ZID": [["31000", "1.3", "0.0"]]}
        v = utils.parse_zid_as_float(h)
        self.assertEqual(v, 1.3)

        s = r"""#TESTID = B38C2094
            #XYID = 31000,108025,432470
            #ZID = 31000,-1.5
            #MEASUREMENTTEXT = 9, maaiveld, vast horizontaal niveau"""
        v = utils.parse_zid_as_float(s)
        self.assertEqual(v, -1.5)

        h = {
            "TESTID": [["B38C2094"]],
            "XYID": [["31000", "108025", "432470"]],
            "ZID": [["31000", "-1.5"]],
            "MEASUREMENTTEXT": [["9", "maaiveld", "vast horizontaal niveau"]],
        }
        v = utils.parse_zid_as_float(h)
        self.assertEqual(v, -1.5)

    def test_cone_id(self):
        s = "#MEASUREMENTTEXT= 4, CFI, conus type"
        v = utils.parse_cone_id(s)
        self.assertEqual("CFI", v)

        h = {"MEASUREMENTTEXT": [["4", "CFI", "conus type"]]}
        v = utils.parse_cone_id(h)
        self.assertEqual("CFI", v)

        s = "#MEASUREMENTTEXT = 4, C10CFIIP.C18469, cone type and serial number"
        v = utils.parse_cone_id(s)
        self.assertEqual("C10CFIIP.C18469", v)

        h = {
            "MEASUREMENTTEXT": [["4", "C10CFIIP.C18469", "cone type and serial number"]]
        }
        v = utils.parse_cone_id(h)
        self.assertEqual("C10CFIIP.C18469", v)

        s = "#MEASUREMENTTEXT=4, S15CFII.d82, Cone Type"
        v = utils.parse_cone_id(s)
        self.assertEqual("S15CFII.d82", v)

        h = {"MEASUREMENTTEXT": [["4", "S15CFII.d82", "Cone Type"]]}
        v = utils.parse_cone_id(h)
        self.assertEqual("S15CFII.d82", v)

    def test_parse_test_id(self):
        # Test ID without spaces on a single line
        s = "#TESTID= CPT-01"
        v = utils.parse_test_id(s)
        self.assertEqual("CPT-01", v)

        h = {"TESTID": [["CPT-01"]]}
        v = utils.parse_test_id(h)
        self.assertEqual("CPT-01", v)

        # Test ID without spaces on a single line and trailing white space
        s = "#TESTID= CPT-02   "
        v = utils.parse_test_id(s)
        self.assertEqual("CPT-02", v)

        h = {"TESTID": [["CPT-02"]]}
        v = utils.parse_test_id(h)
        self.assertEqual("CPT-02", v)

        # Test ID with a space on a single line
        s = "#TESTID= CPT 03"
        v = utils.parse_test_id(s)
        self.assertEqual("CPT 03", v)

        h = {"TESTID": [["CPT 03"]]}
        v = utils.parse_test_id(h)
        self.assertEqual("CPT 03", v)

        # Test ID with a space and trailing whitespace
        s = "#TESTID= CPT 03   "
        v = utils.parse_test_id(s)
        self.assertEqual("CPT 03", v)

        h = {"TESTID": [["CPT 03"]]}
        v = utils.parse_test_id(h)
        self.assertEqual("CPT 03", v)

        # Test ID with 2 spaces in the name
        s = "#TESTID= CPTU17.7 + 83BITE"
        v = utils.parse_test_id(s)
        self.assertEqual("CPTU17.7 + 83BITE", v)

        h = {"TESTID": [["CPTU17.7 + 83BITE"]]}
        v = utils.parse_test_id(h)
        self.assertEqual("CPTU17.7 + 83BITE", v)

    def test_parse_gef_type(self):
        s = r"#PROCEDURECODE= GEF-CPT-Report"
        v = utils.parse_gef_type(s)
        self.assertEqual(v, "cpt")

        h = {"PROCEDURECODE": [["GEF-CPT-Report"]]}
        v = utils.parse_gef_type(h)
        self.assertEqual(v, "cpt")

    def test_xyid(self):
        s = r"#XYID= 31000, 132127.181, 458102.351, 0.000, 0.000"
        x = utils.parse_xid_as_float(s)
        y = utils.parse_yid_as_float(s)
        self.assertEqual(x, 132127.181)
        self.assertEqual(y, 458102.351)

        h = {"XYID": [["31000", "132127.181", "458102.351", "0.000", "0.000"]]}
        x = utils.parse_xid_as_float(h)
        y = utils.parse_yid_as_float(h)
        self.assertEqual(x, 132127.181)
        self.assertEqual(y, 458102.351)

    def test_columns_number(self):
        s = (
            r"#COLUMNINFO = 1, m, Diepte bovenkant laag, 1 #COLUMNINFO = 2, m, Diepte onderkant laag, 2"
            r"#COLUMNINFO = 3, mm, Zandmediaan, 8"
            r"#COLUMNINFO = 4, mm, Grindmediaan, 9"
        )
        v = utils.parse_columns_number(s)
        self.assertEqual(v, 4)

    def test_quantity_number(self):
        s = r"#COLUMNINFO= 1, m, Sondeerlengte, 1"
        v = utils.parse_quantity_number(s, 1)
        self.assertEqual(v, 1)

        h = {"COLUMNINFO": [["1", "m", "Sondeerlengte", "1"]]}
        v = utils.parse_quantity_number(h, 1)
        self.assertEqual(v, 1)

        s = r"#COLUMNINFO= 7, m, Gecorrigeerde diepte, 11"
        v = utils.parse_quantity_number(s, 7)
        self.assertEqual(v, 11)

        h = {"COLUMNINFO": [["7", "m", "Gecorrigeerde diepte", "11"]]}
        v = utils.parse_quantity_number(h, 7)
        self.assertEqual(v, 11)

        s = r"#COLUMNINFO= 4, %, Wrijvingsgetal Rf, 4"
        v = utils.parse_quantity_number(s, 4)
        self.assertEqual(v, 4)

        h = {"COLUMNINFO": [["4", "%", "Wrijvingsgetal Rf", "4"]]}
        v = utils.parse_quantity_number(h, 4)
        self.assertEqual(v, 4)

        s = r"#COLUMNINFO= 4, Graden(deg), Helling, 8"
        v = utils.parse_quantity_number(s, 4)
        self.assertEqual(v, 8)

        h = {"COLUMNINFO": [["4", "Graden(deg)", "Helling", "8"]]}
        v = utils.parse_quantity_number(h, 4)
        self.assertEqual(v, 8)

    def test_column_info(self):
        s = r"#COLUMNINFO= 1, m, Sondeerlengte, 1"
        v = utils.parse_column_info(s, 1, MAP_QUANTITY_NUMBER_COLUMN_NAME_CPT)
        self.assertEqual(v, "penetration_length")

        h = {"COLUMNINFO": [["1", "m", "Sondeerlengte", "1"]]}
        v = utils.parse_column_info(h, 1, MAP_QUANTITY_NUMBER_COLUMN_NAME_CPT)
        self.assertEqual(v, "penetration_length")

    def test_end_of_the_header(self):
        s = r"#EOH="
        v = utils.parse_end_of_header(s)
        self.assertEqual(v, "#EOH=")

    def test_parse_data(self):
        header_s = "This is an header"
        df = pl.DataFrame({"col1": [1, 2, 3], "col2": [1, 2, 3], "col3": [1, 2, 3]})
        data_s = "\n1,1,1\n2,2,2\n3,3,3\n".replace(",", " ")
        df_parsed = _GefCpt.parse_data(
            header_s, data_s, column_names=["col1", "col2", "col3"]
        )
        assert df_parsed.frame_equal(df, null_equal=True)

        # Test terribly formatted columns
        data_s = "\n 1  1 1 \n2 2 2  \n3 3 3\n"
        df_parsed = _GefCpt.parse_data(
            header_s, data_s, column_names=["col1", "col2", "col3"]
        )
        assert df_parsed.frame_equal(df, null_equal=True)

    def test_parse_column_separator(self):
        s = r"#COLUMNSEPARATOR = ;"
        v = utils.parse_column_separator(s)
        self.assertEqual(v, ";")

        h = {"COLUMNSEPARATOR": [[";"]]}
        v = utils.parse_column_separator(h)
        self.assertEqual(v, ";")

    def test_parse_record_separator(self):
        s = r"#RECORDSEPARATOR = !"
        v = utils.parse_record_separator(s)
        self.assertEqual(v, "!")

        h = {"RECORDSEPARATOR": [["!"]]}
        v = utils.parse_record_separator(h)
        self.assertEqual(v, "!")

    def test_find_separator(self):
        s = r"#COLUMNSEPARATOR = ;"
        v = utils.find_separator(s)
        self.assertEqual(v, ";")

        h = {"COLUMNSEPARATOR": [[";"]]}
        v = utils.find_separator(h)
        self.assertEqual(v, ";")

        s = r"I'm sorry the column separator is not in this gef file, even if he wanted to be there."
        v = utils.find_separator(s)
        self.assertEqual(v, r" ")

    def test_create_soil_type(self):
        s = "'Kz'"
        v = utils.create_soil_type(s)
        self.assertEqual(v, "clay 100% with sand")

    def test_parse_data_column_info(self):
        header_s = "This is an header"
        df = pl.DataFrame({"col1": [1, 2, 3], "col2": [1, 2, 3], "col3": [1, 2, 3]})
        data_s = "\n1;1;1\n2;2;2\n3;3;3\n"
        sep = ";"
        df_parsed = _GefBore.parse_data_column_info(
            header_s, data_s, sep, 3, columns_info=["col1", "col2", "col3"]
        )
        assert df_parsed.frame_equal(df, null_equal=True)

    def test_parse_data_soil_type(self):
        df = pl.DataFrame(
            {
                "Soil_type": [
                    "clay 100% with sand",
                    "clay 95% with sand 5%",
                    "clay 90% with sand 10%",
                ]
            }
        )
        data_s = [["'Kz'", "''"], ["'Kz1'", "''"], ["'Kz2'", "''"]]
        df_parsed = _GefBore.parse_data_soil_type(pl.DataFrame({}), data_s)
        assert df_parsed.frame_equal(df, null_equal=True)

    def test_parse_add_info(self):
        s = "'SCH1'"
        v = utils.parse_add_info(s)
        self.assertEqual(v, "1) spoor schelpmateriaal <1% ")

        s = "'DO TOL RO'"
        v = utils.parse_add_info(s)
        self.assertEqual(v, "1) dark olive-red ")

        s = "'BIO'"
        v = utils.parse_add_info(s)
        self.assertEqual(v, "1) bioturbatie ")

        s = "'KEL DR'"
        v = utils.parse_add_info(s)
        self.assertEqual(v, "1) keileem Formatie van Drente ")

    def test_parse_add_info_as_string(self):
        df = pl.DataFrame(
            {
                "remarks": [
                    "1) spoor schelpmateriaal <1% ",
                    "1) dark olive-red ",
                    "1) keileem Formatie van Drente ",
                ]
            }
        )
        data_s = [
            ["'Kz'", "'SCH1'", "''"],
            ["'Kz1'", "'DO TOL RO'", "''"],
            ["'Kz2'", "'KEL DR'", "''"],
        ]

        df_parsed = _GefBore.parse_add_info_as_string(pl.DataFrame({}), data_s)
        assert df_parsed.frame_equal(df, null_equal=True)

    def test_soil_quantification(self):
        s = "'Kz'"
        v = utils.soil_quantification(s)
        self.assertTrue(np.all(np.isclose(v, [0, 0.05, 0.95, 0.0, 0.0, 0.0])))

        s = "'Kz1'"
        v = utils.soil_quantification(s)
        self.assertTrue(np.all(np.isclose(v, [0.0, 0.05, 0.95, 0.0, 0.0, 0.0])))

        s = "'Kz1s1'"
        v = utils.soil_quantification(s)
        self.assertTrue(np.all(np.isclose(v, [0.0, 0.05, 0.9, 0.0, 0.0, 0.05])))

    def test_parse_data_soil_code(self):
        df = pl.DataFrame({"Soil_code": ["Kz", "Kz1", "Kz2"]})
        data_s = [["'Kz'", "''"], ["'Kz1'", "''"], ["'Kz2'", "''"]]
        df_parsed = _GefBore.parse_data_soil_code(pl.DataFrame({}), data_s)
        assert df_parsed.frame_equal(df, null_equal=True)

    def test_data_soil_quantified(self):
        lst = [[0.0, 0.05, 0.95, 0.0, 0.0, 0.0], [0.0, 0.05, 0.95, 0.0, 0.0, 0.0]]
        df = pl.DataFrame(
            lst, columns=["Gravel", "Sand", "Clay", "Loam", "Peat", "Silt"]
        )
        data_s = [["'Kz'", "''"], ["'Kz1'", "''"]]
        df_parsed = _GefBore.parse_soil_quantification(pl.DataFrame({}), data_s)
        assert df_parsed.frame_equal(df, null_equal=True)

    def test_calculate_elevation_respect_to_NAP(self):
        df1 = pl.DataFrame({"depth": [0.0, 1.0, 2.0, 3.0, 4.0]})
        zid = -3
        df_calculated = df1.with_column(
            _GefCpt.calculate_elevation_with_respect_to_nap(zid, 31000)
        )
        df = pl.DataFrame(
            {
                "depth": [0.0, 1.0, 2.0, 3.0, 4.0],
                "elevation_with_respect_to_nap": [-3.0, -4.0, -5.0, -6.0, -7.0],
            }
        )
        assert df_calculated.frame_equal(df, null_equal=True)

    def test_correct_depth_with_inclination(self):
        df1 = pl.DataFrame({"penetration_length": [0.0, 0.2, 0.4, 0.6, 0.8]})
        df_calculated = df1.with_column(correct_depth_with_inclination(df1.columns))
        df = pl.DataFrame(
            {
                "penetration_length": [0.0, 0.2, 0.4, 0.6, 0.8],
                "depth": [0.0, 0.2, 0.4, 0.6, 0.8],
            }
        )
        assert df_calculated.frame_equal(df, null_equal=True)

        df2 = pl.DataFrame(
            {
                "penetration_length": [0.0, 0.2, 0.4, 0.6, 0.8],
                "inclination": [45, 45, 45, 45, 45],
            }
        )
        df_calculated = df2.with_column(correct_depth_with_inclination(df2.columns))
        df = pl.DataFrame(
            {
                "penetration_length": [0.0, 0.2, 0.4, 0.6, 0.8],
                "inclination": [45, 45, 45, 45, 45],
                "depth": [
                    0.0,
                    0.14142135623730953,
                    0.28284271247461906,
                    0.42426406871192857,
                    0.5656854249492381,
                ],
            }
        )
        assert np.isclose(df_calculated["depth"], df["depth"]).all()

        df2 = pl.DataFrame(
            {
                "penetration_length": [0.0, 0.2, 0.4, 0.6, 0.8],
                "corrected_depth": [0.0, 0.10, 0.25, 0.40, 0.60],
                "inclination": [45, 45, 45, 45, 45],
            }
        )
        df_calculated = df2.with_column(correct_depth_with_inclination(df2.columns))
        df = pl.DataFrame(
            {
                "penetration_length": [0.0, 0.2, 0.4, 0.6, 0.8],
                "depth": [0.0, 0.10, 0.25, 0.40, 0.60],
                "inclination": [45, 45, 45, 45, 45],
            }
        )
        assert np.isclose(df_calculated["depth"], df["depth"]).all()

    def test_pre_excavated_depth(self):
        df1 = pl.DataFrame(
            {
                "penetration_length": [0.0, 1.0, 2.0, 3.0, 4.0],
                "qc": [0.5, 0.5, 0.6, 0.7, 0.8],
            }
        )
        pre_excavated_depth = 2
        df_calculated = correct_pre_excavated_depth(df1, pre_excavated_depth)
        df = pl.DataFrame(
            {"penetration_length": [2.0, 3.0, 4.0], "qc": [0.6, 0.7, 0.8]}
        )
        assert df_calculated.frame_equal(df, null_equal=True)

    def test_replace_column_void(self):
        column_void = 999.0
        df1 = pl.DataFrame(
            {
                "penetration_length": [0.0, 1.0, 2.0, 3.0, 4.0],
                "qc": [column_void, 0.5, 0.6, 0.7, 0.8],
            }
        )
        df_calculated = replace_column_void(df1, column_void)
        df = pl.DataFrame(
            {
                "penetration_length": [1.0, 2.0, 3.0, 4.0],
                "qc": [0.5, 0.6, 0.7, 0.8],
            }
        )
        assert df_calculated.frame_equal(df, null_equal=True)

        df1 = pl.DataFrame(
            {
                "penetration_length": [1.0, 2.0, 3.0, 4.0],
                "qc": [0.5, 0.6, column_void, 0.8],
            }
        )

        df_calculated = replace_column_void(df1, column_void)
        assert df_calculated.frame_equal(df, null_equal=True)

    def test_calculate_friction_number(self):
        df1 = pl.DataFrame(
            {"qc": [0.5, 0.5, 0.6, 0.7, 0.8], "fs": [0.0, 0.05, 0.06, 0.07, 0.08]}
        )
        df_calculated = df1.with_column(calculate_friction_number(df1.columns))
        df = pl.DataFrame(
            {
                "qc": [0.5, 0.5, 0.6, 0.7, 0.8],
                "fs": [0.0, 0.05, 0.06, 0.07, 0.08],
                "friction_number": [0.0, 10.0, 10.0, 10.000000000000002, 10.0],
            }
        )
        assert df_calculated.frame_equal(df, null_equal=True)

    def test_parse_cpt(self):
        cpt = Cpt(
            content=dict(
                string="""
#GEFID= 1, 1, 0
#FILEOWNER= Wagen 2
#FILEDATE= 2004, 1, 14
#PROJECTID= CPT, 146203
#COLUMN= 3
#COLUMNINFO= 1, m, Sondeerlengte, 1
#COLUMNINFO= 2, MPa, Conuswaarde, 2
#COLUMNINFO= 3, MPa, Wrijvingsweerstand, 3
#XYID= 31000, 132127.181, 458102.351, 0.000, 0.000
#ZID= 31000, 1.3, 0.0
#MEASUREMENTTEXT= 4, 030919, Conusnummer
#MEASUREMENTTEXT= 6, NEN 5140, Norm
#MEASUREMENTTEXT= 9, MV, fixed horizontal level
#MEASUREMENTVAR= 20, 0.000000, MPa, Nulpunt conus voor sondering
#MEASUREMENTVAR= 22, 0.000000, MPa, Nulpunt kleef voor sondering
#MEASUREMENTVAR= 30, 0.000000, deg, Nulpunt helling voor sondering
#PROCEDURECODE= GEF-CPT-Report, 1, 0, 0, -
#TESTID= 4
#PROJECTNAME= Uitbreiding Rijksweg 2
#OS= DOS
#EOH=
0.0000e+000 0.0000e+000 0.0000e+000
1.0200e+000 7.1000e-001 4.6500e-002
1.0400e+000 7.3000e-001 4.2750e-002
1.0600e+000 6.9000e-001 3.9000e-002
""",
                file_type="gef",
            )
        )
        df_calculated = cpt.df
        df = pl.DataFrame(
            {
                "penetration_length": [0.0000e000, 1.0200e000, 1.0400e000, 1.0600e000],
                "qc": [0.0000e000, 7.1000e-001, 7.3000e-001, 6.9000e-001],
                "fs": [0.0000e000, 4.6500e-002, 4.2750e-002, 3.9000e-002],
                "depth": [0.0000e000, 1.0200e000, 1.0400e000, 1.0600e000],
                "elevation_with_respect_to_nap": [1.3, 0.28, 0.26, 0.24],
                "friction_number": [np.nan, 6.54929577, 5.85616438, 5.65217391],
            }
        )

        # TODO: replace with frame_equal when rounding is supported
        # assert df_calculated.frame_equal(df, null_equal=True)
        for column in df.columns:
            assert df_calculated[column].round(6) == df[column]

    def test_parse_bore(self):
        cpt = Bore(
            content=dict(
                string="""
#GEFID = 1,1,0
#COLUMNTEXT = 1, aan
#COLUMNSEPARATOR = ;
#RECORDSEPARATOR = !
#FILEOWNER = DINO
#FILEDATE = 2010,9,1
#PROJECTID = DINO-BOR
#COLUMN = 9
#COLUMNINFO = 1, m, Diepte bovenkant laag, 1
#COLUMNINFO = 2, m, Diepte onderkant laag, 2
#COLUMNINFO = 3, mm, Zandmediaan, 8
#COLUMNINFO = 4, mm, Grindmediaan, 9
#COLUMNINFO = 5, %, Lutum percentage, 3
#COLUMNINFO = 6, %, Silt percentage, 4
#COLUMNINFO = 7, %, Zand percentage, 5
#COLUMNINFO = 8, %, Grind percentage, 6
#COLUMNINFO = 9, %, Organische stof percentage, 7
#COLUMNVOID = 1, -9999.99
#COLUMNVOID = 2, -9999.99
#COLUMNVOID = 3, -9999.99
#COLUMNVOID = 4, -9999.99
#COLUMNVOID = 5, -9999.99
#COLUMNVOID = 6, -9999.99
#COLUMNVOID = 7, -9999.99
#COLUMNVOID = 8, -9999.99
#COLUMNVOID = 9, -9999.99
#LASTSCAN = 44
#REPORTCODE = GEF-BORE-Report,1,0,0
#MEASUREMENTCODE = Onbekend
#TESTID = B25G0304
#XYID = 31000,120870,483400
#ZID = 31000,2.0
#MEASUREMENTVAR = 19, 1, -, aantal peilbuizen
#EOH =
0.00;1.20;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Zgh2';'TGR GE';'ZMFO';'CA3';!
1.20;3.10;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Zg';'ON';'ZMGO';'FN2';'CA2';!
3.10;5.00;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Vz';'TBR ZW';'ZMO';'CA1';!
""",
                file_type="gef",
            )
        )
        df_calculated = cpt.df
        df = pl.DataFrame(
            {
                "depth_top": [0.0, 1.2, 3.1],
                "depth_bottom": [1.2, 3.1, 5.0],
                "soil_code": ["Zgh2", "Zg", "Vz"],
                "remarks": [
                    "1) gray-yellow 2) ZMFO 3) kalkrijk ",
                    "1) ON 2) ZMGO 3) weinig fijn grind (1-25%) 4) kalkarm ",
                    "1) brown-black 2) ZMO 3) kalkloos ",
                ],
                "g": [0.05, 0.05, 0.00],
                "s": [0.85, 0.95, 0.05],
                "c": [0.0, 0.0, 0.0],
                "l": [0.0, 0.0, 0.0],
                "p": [0.10, 0.00, 0.95],
                "si": [0.0, 0.0, 0.0],
            }
        )
        assert df_calculated.frame_equal(df, null_equal=True)

    def test_parse_pre_excavated_dept_with_void_inclination(self):
        cpt = Cpt(
            content=dict(
                string="""
#COLUMN= 6
#COLUMNINFO= 1, m, Sondeerlengte, 1
#COLUMNINFO= 2, MPa, Conuswaarde, 2
#COLUMNINFO= 3, MPa, Wrijvingsweerstand, 3
#COLUMNINFO= 4, Deg, Helling, 8
#COLUMNINFO= 5, %, Wrijvingsgetal, 4
#COLUMNINFO= 6, MPa, Waterspanning, 5
#COLUMNVOID= 2, -9999.000000
#COLUMNVOID= 3, -9999.000000
#COLUMNVOID= 4, -9999.000000
#COLUMNVOID= 5, -9999.000000
#COLUMNVOID= 6, -9999.000000
#LASTSCAN= 3
#ZID= 31000, -0.39, 0.05
#MEASUREMENTVAR= 13, 1.500000, m, voorgeboorde/voorgegraven diepte
#REPORTCODE= GEF-CPT-Report, 1, 1, 0, -
#EOH=
0.0000e+000 -9.9990e+003 -9.9990e+003 -9.9990e+003 -9.9990e+003 -9.9990e+003
1.5100e+000 9.1800e+000 5.3238e-002 5.8398e-001 5.7314e-001 3.0107e-003
1.5300e+000 9.3044e+000 5.3803e-002 8.2007e-001 5.7986e-001 3.3362e-003
""",
                file_type="gef",
            )
        )
        expected = pl.DataFrame(
            {
                "penetration_length": [1.51, 1.53],
                "qc": [9.1800, 9.3044],
                "fs": [0.053238, 0.053803],
                "inclination": [0.58398, 0.82007],
                "friction_number": [0.579935, 0.578253],
                "u1": [0.003011, 0.003336],
                "depth": [1.510000, 1.529999],
                "elevation_with_respect_to_nap": [-1.90, -1.919999],
            }
        )
        # TODO: replace with frame_equal when rounding is supported
        # assert cpt.df.frame_equal(expected, null_equal=True)
        for column in cpt.df.columns:
            assert expected[column].round(6) == cpt.df[column]

    def test_delta_depth(self):
        df1 = pl.DataFrame({"depth": [0.0, 0.5, 1.0]})
        v = geo.delta_depth(df1)
        df = pl.DataFrame({"depth": [0.0, 0.5, 1.0], "delta_depth": [0.0, 0.5, 0.5]})
        assert v.frame_equal(df, null_equal=True)

    def test_soil_pressure(self):
        v = pl.DataFrame({"gamma": [0.0, 0.5, 1.0], "delta_depth": [0.0, 0.5, 0.5]})
        geo.soil_pressure(v)
        df = pl.DataFrame(
            {
                "gamma": [0.0, 0.5, 1.0],
                "delta_depth": [0.0, 0.5, 0.5],
                "soil_pressure": [0.0, 0.25, 0.75],
            }
        )
        assert v.frame_equal(df, null_equal=True)

    def test_water_pressure(self):
        water_level = 0.5
        v = pl.DataFrame({"depth": [0.0, 0.5, 1.0]})
        geo.water_pressure(v, water_level)
        df = pl.DataFrame(
            {"depth": [0.0, 0.5, 1.0], "water_pressure": [0.0, 0.0, 4.905]}
        )

        assert v.frame_equal(df, null_equal=True)

    def test_effective_soil_pressure(self):
        df1 = pl.DataFrame(
            {"soil_pressure": [0.0, 0.25, 0.75], "water_pressure": [0.0, 0.0, 4.905]}
        )
        v = geo.effective_soil_pressure(df1)
        df = pl.DataFrame(
            {
                "soil_pressure": [0.0, 0.25, 0.75],
                "water_pressure": [0.0, 0.0, 4.905],
                "effective_soil_pressure": [0.0, 0.25, -4.155],
            }
        )
        assert v.frame_equal(df, null_equal=True)

    def test_assign_multiple_columns(self):
        df1 = pl.DataFrame(
            {"soil_pressure": [0.0, 0.25, 0.75], "water_pressure": [0.0, 0.0, 4.905]}
        )
        v = utils.assign_multiple_columns(df1, ["soil_pressure", "water_pressure"], df1)
        df = pl.DataFrame(
            {"soil_pressure": [0.0, 0.25, 0.75], "water_pressure": [0.0, 0.0, 4.905]}
        )
        assert v.frame_equal(df, null_equal=True)

    def test_kpa_to_mpa(self):
        df1 = pl.DataFrame(
            {"soil_pressure": [0.0, 0.25, 0.75], "water_pressure": [0.0, 0.0, 4.905]}
        )
        v = utils.kpa_to_mpa(df1, ["soil_pressure", "water_pressure"])
        df = pl.DataFrame(
            {
                "soil_pressure": [0.0, 0.00025, 0.00075],
                "water_pressure": [0.0, 0.0, 0.004905],
            }
        )
        # TODO: replace with frame_equal when rounding is supported
        # assert df_calculated.frame_equal(df, null_equal=True)
        for column in df.columns:
            assert v[column].round(6) == df[column]

    def test_qt(self):
        df1 = pl.DataFrame({"qc": [0.0, 1.0, 2.0], "u2": [0.0, 1.0, 1.0]})
        v = geo.qt(df1, area_quotient_cone_tip=0.5)
        df = pl.DataFrame(
            {"qc": [0.0, 1.0, 2.0], "u2": [0.0, 1.0, 1.0], "qt": [0.0, 1.5, 2.5]}
        )
        assert v.frame_equal(df, null_equal=True)

    def test_normalized_cone_resistance(self):
        df1 = pl.DataFrame(
            {
                "qt": [0.0, 1.5, 2.5],
                "soil_pressure": [0.0, 0.25, 0.75],
                "effective_soil_pressure": [0.0, 0.25, -4.155],
            }
        )
        v = geo.normalized_cone_resistance(df1)
        df = pl.DataFrame(
            {
                "qt": [0.0, 1.5, 2.5],
                "soil_pressure": [0.0, 0.25, 0.75],
                "effective_soil_pressure": [0.0, 0.25, -4.155],
                "normalized_cone_resistance": [-np.nan, 5.0, 1.0],
            }
        )
        # TODO: why aren't these equal with df.frame_equal?
        assert v.to_csv() == df.to_csv()

    def test_normalized_friction_ratio(self):
        df1 = pl.DataFrame(
            {
                "qt": [0.0, 1.5, 2.5],
                "fs": [0.5, 0.5, 0.5],
                "soil_pressure": [0.0, 0.25, 0.75],
                "effective_soil_pressure": [0.0, 0.25, -4.155],
            }
        )
        v = geo.normalized_friction_ratio(df1)
        df = pl.DataFrame(
            {
                "qt": [0.0, 1.5, 2.5],
                "fs": [0.5, 0.5, 0.5],
                "soil_pressure": [0.0, 0.25, 0.75],
                "effective_soil_pressure": [0.0, 0.25, -4.155],
                "normalized_friction_ratio": [np.inf, 40.0, 28.57142857142857],
            }
        )
        assert v.frame_equal(df, null_equal=True)

    def test_nan_to_zero(self):
        df1 = pl.DataFrame({"type_index": [None, 1.0, 2.0]})
        v = utils.none_to_zero(df1)
        df = pl.DataFrame({"type_index": [0.0, 1.0, 2.0]})
        assert v.frame_equal(df, null_equal=True)

    def test_bug_depth(self):
        cpt = """
#FILEDATE= 2011, 5, 13
#PROJECTID= CPT, 4015110
#COLUMN= 5
#COLUMNINFO= 1, m, Sondeerlengte, 1
#COLUMNINFO= 2, MPa, Conuswaarde, 2
#COLUMNINFO= 3, MPa, Wrijvingsweerstand, 3
#COLUMNINFO= 4, Deg, Helling, 8
#COLUMNINFO= 5, %, Wrijvingsgetal, 4
#XYID= 31000, 116371.70, 514039.34, 0.02, 0.02
#ZID= 31000, -2.21, 0.05
#MEASUREMENTTEXT= 4, S10CFI518, conusnummer
#MEASUREMENTTEXT= 6, NEN 5140 klasse2, gehanteerde norm en klasse sondering
#MEASUREMENTTEXT= 9, maaiveld, vast horizontaal vlak
#MEASUREMENTVAR= 1, 1000.000000, mm2, nom. oppervlak conuspunt
#MEASUREMENTVAR= 12, 0.000000, -, elektrische sondering
#MEASUREMENTVAR= 13, 1.500000, m, voorgeboorde/voorgegraven diepte
#TESTID= 29
#PROJECTNAME= Herinrichting van de N243
#REPORTCODE= GEF-CPT-Report, 1, 1, 0, -
#REPORTTEXT= 201, Mos Grondmechanica B.V.
#REPORTTEXT= 202, Herinrichting van de N243 te Schermer en Beemster
#REPORTTEXT= 203, Sondering 29
#STARTDATE= 2010, 12, 7
#STARTTIME= 9, 0, 21.000000
#OS= DOS
#EOH=
0.0000e+000 -9.9990e+003 -9.9990e+003 -9.9990e+003 -9.9990e+003
1.5100e+000 3.6954e-001 3.7981e-003 9.9502e-002 9.2308e-001
1.5300e+000 4.0370e-001 5.5589e-003 1.1194e-001 1.3260e+000
1.5500e+000 4.2854e-001 7.5360e-003 1.1194e-001 1.8315e+000
1.5700e+000 4.4407e-001 1.0028e-002 1.1194e-001 2.5117e+000
1.5900e+000 4.5028e-001 9.4163e-003 9.9502e-002 2.4202e+000
1.6100e+000 3.7265e-001 8.6356e-003 9.9502e-002 2.3201e+000
1.6300e+000 3.2607e-001 7.9297e-003 1.1194e-001 2.2770e+000
1.6500e+000 2.9812e-001 7.9607e-003 9.9502e-002 2.4315e+000
1.6700e+000 2.8570e-001 7.0938e-003 1.2438e-001 2.2649e+000
1.6900e+000 2.6085e-001 6.4691e-003 1.1194e-001 1.9653e+000
1.7100e+000 2.9812e-001 5.9068e-003 9.9502e-002 1.6748e+000
1.7300e+000 3.5091e-001 5.2480e-003 9.9502e-002 1.4391e+000
1.7500e+000 4.8444e-001 3.8332e-003 9.9502e-002 1.0214e+000
1.7700e+000 4.9065e-001 3.5531e-003 9.9502e-002 9.0498e-001
1.7900e+000 3.8196e-001 3.4469e-003 1.1194e-001 8.5759e-001
1.8100e+000 3.6023e-001 4.1928e-003 9.9502e-002 1.0363e+000
1.8300e+000 3.8196e-001 4.5735e-003 1.1194e-001 1.1782e+000
1.8500e+000 3.6333e-001 5.2490e-003 1.1194e-001 1.4342e+000
1.8700e+000 3.6954e-001 5.2590e-003 9.9502e-002 1.4635e+000
1.8900e+000 3.6954e-001 5.8176e-003 1.1194e-001 1.5781e+000
        """
        cpt = Cpt(content=dict(string=cpt, file_type="gef"))
        assert np.isclose(cpt.df[0, "depth"], 1.51)

    def test_bore_with_reduced_columns(self):
        Bore(
            content=dict(
                string="""
#PROJECTID= redacted, 1234, -
#COLUMN= 2
#COLUMNINFO= 1, m, Laag van, 1
#COLUMNINFO= 2, m, Laag tot, 2
#DATAFORMAT= ASCII
#COLUMNSEPARATOR= ;
#COLUMNTEXT= 1
#PROCEDURECODE= GEF-BORE-Report, 1, 0, 0, -
#REPORTCODE= GEF-BORE-Report, 1, 0, 0, -
#RECORDSEPARATOR= !
#OS= DOS
#LANGUAGE= NL
#EOH=
0.0000e+000;2.0000e-001;'Zs1';'ZUF';'DO TGR BR';;;!
2.0000e-001;4.0000e-001;'Vz3';;'DO BR';;;!
4.0000e-001;6.0000e-001;'Zs1';'ZUF';'DO TBR GR';;;!
6.0000e-001;1.0000e+000;'Zs1';'ZUF';'LI BR';;;!
1.0000e+000;1.4000e+000;'Vk1';;'DO TRO BR';;;!
1.4000e+000;2.2000e+000;'Zs1';'ZUF';'LI BR';'Restante BZB.: PR (zwak)';;!
2.2000e+000;2.6000e+000;'Zs1g1';'ZUF';'LI TBE BR';'Restante BZB.: PR (zwak)';;!
2.6000e+000;3.2000e+000;'Zs1g1';'ZUF';'LI GR';;;!
""",
                file_type="gef",
            )
        )


class BoreTest(unittest.TestCase):
    def setUp(self):
        self.bore = Bore(
            content=dict(
                string="""
#GEFID = 1,1,0
#COLUMNTEXT = 1, aan
#COLUMNSEPARATOR = ;
#RECORDSEPARATOR = !
#FILEOWNER = DINO
#COMPANYID = Wiertsema & Partners
#FILEDATE = 2015,7,15
#PROJECTID = DINO-BOR
#COLUMN = 9
#COLUMNINFO = 1, m, Diepte bovenkant laag, 1
#COLUMNINFO = 2, m, Diepte onderkant laag, 2
#COLUMNINFO = 3, mm, Zandmediaan, 8
#COLUMNINFO = 4, mm, Grindmediaan, 9
#COLUMNINFO = 5, %, Lutum percentage, 3
#COLUMNINFO = 6, %, Silt percentage, 4
#COLUMNINFO = 7, %, Zand percentage, 5
#COLUMNINFO = 8, %, Grind percentage, 6
#COLUMNINFO = 9, %, Organische stof percentage, 7
#COLUMNVOID = 1, -9999.99
#COLUMNVOID = 2, -9999.99
#COLUMNVOID = 3, -9999.99
#COLUMNVOID = 4, -9999.99
#COLUMNVOID = 5, -9999.99
#COLUMNVOID = 6, -9999.99
#COLUMNVOID = 7, -9999.99
#COLUMNVOID = 8, -9999.99
#COLUMNVOID = 9, -9999.99
#LASTSCAN = 29
#REPORTCODE = GEF-BORE-Report,1,0,0
#MEASUREMENTCODE = Volgens GEF-BORE-Report 1.0.0,-,-,-
#TESTID = B43F1303
#XYID = 31000,99046.00,424271.00
#ZID = 31000,1.96,5.0000002E-5
#MEASUREMENTTEXT = 3, Puttershoek, plaatsnaam
#MEASUREMENTTEXT = 5, 2014-05-07, datum boorbeschrijving
#MEASUREMENTTEXT = 6, Jan Palsma, beschrijver lagen
#MEASUREMENTTEXT = 13, Wiertsema & Partners, boorbedrijf
#MEASUREMENTTEXT = 16, 2014-05-07, datum boring
#MEASUREMENTVAR = 31, 28.00, m, diepte onderkant boortraject
#MEASUREMENTVAR = 32, 168, mm, boorbuisdiameter
#MEASUREMENTTEXT = 31, PUM, boormethode
#MEASUREMENTTEXT = 11, MONB, maaiveldhoogtebepaling
#MEASUREMENTTEXT = 12, LONB, plaatsbepalingmethode
#MEASUREMENTTEXT = 7, Rijksdriehoeksmeting, locaal coördinatensysteem
#MEASUREMENTTEXT = 8, Normaal Amsterdams Peil, locaal referentiesysteem
#MEASUREMENTVAR = 16, 28.00, m, einddiepte
#MEASUREMENTTEXT = 1, Antea Group, opdrachtgever
#MEASUREMENTTEXT = 14, Nee, openbaar
#MEASUREMENTTEXT = 18, Nee, peilbuis afwezig
#MEASUREMENTVAR = 18, 1.00, m, grondwaterstand direct na boring
#EOH =
0.00;0.07;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'NBE';'NBE';'betontegel geen monster';!
0.07;0.50;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Zs2';'LI BR';'ZMG';'zeer veel baksteenresten';!
0.50;2.00;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Zs1g1';'LI BR';'ZZG';'GFN';'SCH1';'uiterst grof zandhoudend plaatselijk weinig schelprestjes opgebracht';!
2.00;3.20;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Zs1';'LI TGR GR';'ZMG';'SCH1';'grof zandhoudend opgebracht';!
3.20;5.10;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Zs2';'LI TGR GR';'ZMG';'SCH1';'plaatselijk weinig kleibrokjes grof zandhoudend opgebracht';!
5.10;6.00;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Kz1';'GR';'KSTV';'SCH1';!
6.00;8.40;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Kz3h1';'GR';'KMST';'zandgelaagd plaatselijk weinig plantenresten rietresten';!
8.40;8.50;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Zs1';'LI GR';'ZMG';!
8.50;8.55;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Kz3';'KMST';'weinig plantenresten houtresten';!
8.55;8.80;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Zs2';'LI TGR GR';'ZMG';'plaatselijk weinig plantenresten houtresten weinig kleibrokjes';!
8.80;9.00;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Ks1h3';'TGR BR';'KSTV';'plaatselijk plantenresten rietresten houtresten';!
9.00;9.90;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Vk1';'DO TBR BR';'VSTV';'houtresten';!
9.90;12.50;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Ks2h1';'TBR GR';'KSTV';'plaatselijk plantenresten rietresten veel houtresten';!
12.50;13.00;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Ks2h1';'TBR GR';'KSTV';'plaatselijk plantenresten rietresten houtresten plaatselijk grove houtresten';!
13.00;13.50;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Ks1h2';'TGR BR';'KSTV';'plaatselijk plantenresten rietresten houtresten';!
13.50;13.90;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Vk3';'DO TBR GR';'VSTV';!
13.90;15.90;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Ks2h1';'GR';'KSTV';'plaatselijk weinig plantenresten rietresten houtresten';!
15.90;16.30;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Vk3';'TGR BR';'VSTV';!
16.30;16.55;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Ks2';'GR';'KSTV';'houtresten';!
16.55;17.45;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Zs2';'LI TGR GR';'ZMG';'plaatselijk plantenresten houtresten rietresten';!
17.45;19.60;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Zs1g1';'LI TGR GR';'ZUG';'GFN';'grof zandhoudend plaatselijk weinig plantenresten rietresten houtresten';!
19.60;22.10;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Zs1g2';'LI TGR GR';'ZUG';'GFN';'uiterst grof/grof zandhoudend';!
22.10;23.20;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Kz2';'TGN GR';'KSTV';!
23.20;23.70;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Kz1';'TGN GR';'KZST';!
23.70;24.20;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Kz3';'GR';'KSTV';!
24.20;24.35;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Zs4';'GR';'ZZF';!
24.35;26.00;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Kz3';'TGN GR';'KSTV';!
26.00;26.80;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Zs3';'LI TGR GR';'ZZF';'plaatselijk kleirestje fijn zandhoudend';!
26.80;28.00;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;-9999.99;'Kz1';'TGN GR';'KZST';!
""",
                file_type="gef",
            )
        )

    def test_sum_to_one(self):
        cols = [
            "gravel_component",
            "sand_component",
            "clay_component",
            "loam_component",
            "peat_component",
            "silt_component",
        ]
        s = self.bore.df.select(
            [
                pl.fold(
                    pl.lit(0),
                    lambda a, b: a + b,
                    [
                        pl.when(pl.col(a) < 0).then(1 / len(cols)).otherwise(pl.col(a))
                        for a in cols
                    ],
                ).alias("sum")
            ]
        )
        self.assertTrue(np.all(np.isclose(s, 1)))


class PlotTest(unittest.TestCase):
    def test_plot_cpt(self):
        gef = Cpt("./pygef/test_files/example.gef")
        gef.plot(show=False)

    def test_plot_bore(self):
        gef = Bore("./pygef/test_files/example_bore.gef")
        gef.plot(show=False, figsize=(4, 12))

    def test_plot_classification(self):
        gef = Cpt("./pygef/test_files/example.gef")
        gef.plot(show=False, classification="robertson", water_level_wrt_depth=-1)

        gef = Cpt("./pygef/test_files/cpt.gef")
        gef.plot(show=False, classification="robertson", water_level_wrt_depth=-1)

        gef = Cpt("./pygef/test_files/cpt2.gef")
        gef.plot(show=False, classification="robertson", water_level_wrt_depth=-1)


class TestRobertson(unittest.TestCase):
    def setUp(self):
        self.gef = Cpt("./pygef/test_files/example.gef")

    def test_nan_dropped(self):
        self.assertAlmostEqual(self.gef.df["qc"][0], 16.72)

    def test_water_pressure(self):
        """
        depth starts at 6 meters, So -7 should lead to water pressure of 0
        """
        df = self.gef.classify(
            "robertson", water_level_NAP=None, water_level_wrt_depth=-7
        )
        self.assertEqual(df["water_pressure"][0], 0)


class TestBoreXml(unittest.TestCase):
    def setUp(self):
        self.bore = Bore("./pygef/test_files/bore_xml/DP65+005_HB_VL.xml")

    def test_bore_main_attributes(self):
        assert np.isclose(self.bore.zid, 4.49)
        assert np.isclose(self.bore.x, 154154.0)
        assert np.isclose(self.bore.y, 443248.0)
        assert self.bore.file_date == datetime(2021, 10, 5).date()
        assert self.bore.test_id == "78379_HB423"

    def test_df_bore(self):

        assert self.bore.df.columns == [
            "depth_top",
            "depth_bottom",
            "soil_name",
            "gravel_component",
            "sand_component",
            "clay_component",
            "loam_component",
            "peat_component",
            "silt_component",
        ]

    def test_all_xml(self):
        path_folder = "./pygef/test_files/bore_xml/"
        for file in os.listdir(path_folder):
            if file.lower().endswith("xml"):
                bore = Bore(f"{path_folder}/{file}")

                assert {
                    "gravel_component",
                    "sand_component",
                    "clay_component",
                    "loam_component",
                    "peat_component",
                    "silt_component",
                }.issubset(set(bore.df.columns))
                bore.plot()

    def test_read_as_string(self):
        path = "./pygef/test_files/bore_xml/DP65+005_HB_VL.xml"
        with open(path, encoding="utf-8", errors="ignore") as f:
            string = f.read()

        bore = Bore(content=dict(string=string, file_type="xml"))
        assert type(bore.s_bin) == bytes
        assert type(bore.s) == str



import numpy as np
import polars as pl
import pygef.utils as utils
from polars import col, lit
from pygef import geo


def excess_pore_pressure_ratio(df):
    """
    Assign the excess pore pressure ratio, if the water pressure u2 is defined. Else, raise ERROR.

    :param df: (DataFrame)
    :return: (DataFrame)
    """
    try:
        u2 = df["u2"]
    except KeyError:
        raise SystemExit("ERROR: u2 not defined in .gef file, change classifier")

    df["excess_pore_pressure_ratio"] = (u2 - df["water_pressure"]) / (
        df["qt"] - df["soil_pressure"]
    )

    return df


def ic_to_gamma(water_level):
    """
    Return the expression needed to compute "gamma_predict"
    """
    below_water = (1.0 - col("depth")) < water_level
    ti = col("type_index")
    return (
        pl.when(ti > 3.22)
        .then(11.0)
        .when((ti <= 3.22) & (ti > 2.76))
        .then(16.0)
        .when((ti <= 2.76) & ~(below_water))
        .then(18.0)
        .when(ti <= 2.40 & below_water)
        .then(19.0)
        .when(ti <= 1.80 & below_water)
        .then(20.0)
        .otherwise(1.0)
        .alias("gamma_predict")
    )


def ic_to_soil_type():
    """
    Assign the soil type to the corresponding Ic.
    """
    ti = col("type_index")
    return (
        pl.when(ti > 3.22)
        .then("Peat")
        .when((ti <= 3.22) & (ti > 2.67))
        .then("Clays")
        .when((ti <= 2.67) & (ti > 2.4))
        .then("Clayey silt to silty clay")
        .when((ti <= 2.4) & (ti > 1.8))
        .then("Silty sand to sandy silt")
        .when((ti <= 1.8) & (ti > 1.25))
        .then("Sands: clean sand to silty")
        .when(ti <= 1.25)
        .then("Gravelly sands")
        .otherwise("")
        .alias("soil_type")
    )


def type_index() -> pl.Expr:
    return (
        (
            (
                pl.lit(3.0)
                - np.log10(
                    col("normalized_cone_resistance")
                    * (1.0 - col("excess_pore_pressure_ratio"))
                    + 1.0
                )
            )
            ** 2
            + (1.5 + 1.3 * np.log10(col("normalized_friction_ratio"))) ** 2
        )
        ** 0.5
    ).alias("type_index")


def iterate_been_jeffrey(
    original_df, water_level, area_quotient_cone_tip=None, pre_excavated_depth=None
):
    """
    Iteration function for Been & Jefferies classifier.

    :param original_df: (DataFrame)
    :param water_level: (int) Water level with respect to ground level.
    :param area_quotient_cone_tip: (float)
    :param pre_excavated_depth: (float)
    :return: (DataFrame)
    """
    gamma = np.ones(original_df.shape[0]) * 18.0

    c = 0
    while True:
        c += 1
        df = original_df
        df["gamma"] = gamma

        def condition(x):
            return np.all(x["gamma_predict"].series_equal(pl.Series(gamma)))

        df = been_jeffrey(
            df,
            water_level,
            area_quotient_cone_tip=area_quotient_cone_tip,
            pre_excavated_depth=pre_excavated_depth,
        )
        df["gamma_predict"] = np.nan_to_num(df["gamma_predict"])
        if condition(df):
            break
        elif c > 10:
            break
        else:
            gamma = df["gamma_predict"]

    return df.with_column(ic_to_soil_type())


def been_jeffrey(
    df, water_level, area_quotient_cone_tip=None, pre_excavated_depth=None
):
    """
    Implementation of Been & Jefferies.

    :param df: (DataFrame)
    :param water_level: (float)
    :param area_quotient_cone_tip: (float)
    :param pre_excavated_depth: (float)
    :return: (DataFrame)
    """
    df = (
        df.with_column(col("depth").diff().alias("delta_depth"))
        .pipe(geo.soil_pressure)
        .pipe(geo.qt, area_quotient_cone_tip)
        .pipe(geo.water_pressure, water_level)
        .pipe(geo.effective_soil_pressure)
        .pipe(
            utils.kpa_to_mpa,
            ["soil_pressure", "effective_soil_pressure", "water_pressure"],
        )
        .pipe(excess_pore_pressure_ratio)
        .pipe(geo.normalized_cone_resistance)
        .pipe(geo.normalized_friction_ratio)
        .pipe(utils.none_to_zero)
        .with_columns([type_index()])
        .with_column(ic_to_gamma(water_level))
    )
    return df



from abc import ABC
from datetime import datetime

import numpy as np
import polars as pl
from lxml import etree

NS_MAP_VALUES = [
    "http://www.broservices.nl/xsd/brocommon/",
    "http://www.broservices.nl/xsd/isbhr-gt/",
    "http://www.opengis.net/gml/",
    "http://www.broservices.nl/xsd/bhrgtcommon/",
]


class _BroXml(ABC):
    def __init__(self, path: str = None, string: str = None):
        """
        Parser of xml files.

        Parameters
        ----------
        path
            Path to the xml file
        string
            Xml file as string
        """
        assert (
            any([path, string]) is not None
        ), "One of [path, string] should be not none."
        self.path = path
        if string is None and path is not None:
            with open(path, "rb") as f:
                self.s_bin = f.read()
            with open(path, encoding="utf-8", errors="ignore") as f:
                string = f.read()
        elif string is not None:
            self.s_bin = string.encode("ascii")

        self.s = string
        # Initialize attributes
        self.zid = None
        self.x = None
        self.y = None
        self.test_id = None
        self.height_system = None


class _BroXmlBore(_BroXml):
    """
    Parser of boreholes xml files.

    Parameters
    ----------
    path
        Path to the xml file
    string
        Xml file as string
    """

    def __init__(self, path=None, string=None):
        super().__init__(path=path, string=string)
        self.nen_version = "NEN-EN-ISO 14688"
        self.type = "bore"
        self._root = etree.fromstring(self.s_bin)
        self._ns_map_keys = list(self._root.nsmap.keys())

        # find versions
        self._ns_map_values = []
        for pos in NS_MAP_VALUES:
            for val in list(self._root.nsmap.values()):
                if pos in val:
                    self._ns_map_values.append(val)
                    break

        self._parse_attributes()

        self.df = self._parse_df()

    def _parse_attributes(self) -> None:
        self.test_id = list(
            self._root.iter(
                "{" + self._ns_map_values[1] + "}" + "objectIdAccountableParty"
            )
        )[0].text
        self.zid = float(
            list(self._root.iter("{" + self._ns_map_values[3] + "}" + "offset"))[0].text
        )
        self.file_date = datetime.strptime(
            list(
                list(
                    self._root.iter("{" + self._ns_map_values[3] + "}boringStartDate")
                )[0].iter("{" + self._ns_map_values[0] + "}" + "date")
            )[0].text,
            "%Y-%m-%d",
        ).date()

        coords = list(self._root.iter("{" + self._ns_map_values[2] + "}pos"))[
            0
        ].text.split()
        self.x = float(coords[0])
        self.y = float(coords[1])

    def _parse_df(self) -> pl.DataFrame:
        depth_top = []
        depth_bottom = []
        soil_name = []
        for layer in self._root.iter("{" + self._ns_map_values[3] + "}layer"):
            depth_top.append(
                float(
                    list(layer.iter("{" + self._ns_map_values[3] + "}upperBoundary"))[
                        0
                    ].text
                )
            )
            depth_bottom.append(
                float(
                    list(layer.iter("{" + self._ns_map_values[3] + "}lowerBoundary"))[
                        0
                    ].text
                )
            )
            for loc in layer:
                if loc.tag == "{" + self._ns_map_values[3] + "}soil":
                    s = list(
                        loc.iter("{" + self._ns_map_values[3] + "}geotechnicalSoilName")
                    )[0].text
                    soil_name.append(s)
                elif loc.tag == "{" + self._ns_map_values[3] + "}specialMaterial":
                    soil_name.append(loc.text)
            if len(soil_name) < len(depth_bottom):
                soil_name.append("not classified")

        df = pl.DataFrame(
            {
                "depth_top": depth_top,
                "depth_bottom": depth_bottom,
                "soil_name": soil_name,
            }
        )
        df = soil_name_to_percentages(df)
        return df


class _BroXmlCpt(_BroXml):
    def __init__(self, path=None, string=None):
        super().__init__(path=path, string=string)

        raise NotImplementedError(
            "The parsing of cpts in xml format is not yet available."
        )


def soil_name_to_percentages(df):
    """
    Adds the columns
    ["gravel_component", "sand_component", "clay_component", "loam_component", "peat_component", "silt_component"]
    to the Dataframe
    Parameters
    ----------
    df
        DataFrame containing 'soil_name' column
    Returns
    -------

    """
    soil_names_dict = {
        "betonOngebroken": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # specialMaterial
        "keitjes": [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        "klei": [0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
        "kleiigVeen": [0.0, 0.0, 0.3, 0.0, 0.7, 0.0],
        "kleiigZand": [0.0, 0.7, 0.3, 0.0, 0.0, 0.0],
        "kleiigZandMetGrind": [0.5, 0.65, 0.3, 0.0, 0.0, 0.0],
        "puin": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # specialMaterial
        "siltigZand": [0.0, 0.7, 0.0, 0.0, 0.0, 0.3],
        "siltigZandMetGrind": [0.5, 0.65, 0.0, 0.0, 0.0, 0.3],
        "sterkGrindigZand": [0.3, 0.7, 0.0, 0.0, 0.0, 0.0],
        "sterkGrindigeKlei": [0.3, 0.0, 0.7, 0.0, 0.0, 0.0],
        "sterkZandigGrind": [0.7, 0.3, 0.0, 0.0, 0.0, 0.0],
        "sterkZandigSilt": [0.0, 0.3, 0.0, 0.0, 0.0, 0.7],
        "sterkZandigeKlei": [0.0, 0.3, 0.7, 0.0, 0.0, 0.0],
        "sterkZandigeKleiMetGrind": [0.5, 0.3, 0.65, 0.0, 0.0, 0.0],
        "veen": [0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
        "zand": [0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
        "zwakGrindigZand": [0.1, 0.9, 0.0, 0.0, 0.0, 0.0],
        "zwakGrindigeKlei": [0.1, 0.0, 0.9, 0.0, 0.0, 0.0],
        "zwakZandigGrind": [0.9, 0.1, 0.0, 0.0, 0.0, 0.0],
        "zwakZandigVeen": [0.0, 0.1, 0.0, 0.0, 0.9, 0.0],
        "zwakZandigeKlei": [0.0, 0.1, 0.9, 0.0, 0.0, 0.0],
        "zwakZandigeKleiMetGrind": [0.5, 0.1, 0.85, 0.0, 0.0, 0.0],
    }

    array_ = np.array(
        [
            soil_names_dict[soil_name]
            if soil_name in soil_names_dict.keys()
            else [0, 0, 0, 0, 0, 0]
            for soil_name in df["soil_name"]
        ]
    )
    df_soils = pl.DataFrame(
        array_,
        columns=[
            "gravel_component",
            "sand_component",
            "clay_component",
            "loam_component",
            "peat_component",
            "silt_component",
        ],
    )
    df = df.hstack(df_soils)
    return df



from abc import ABC


class Base(ABC):
    def __init__(self):
        """
        Abstract base class to parse cpt and boreholes in gef or xml format.

        """
        self.test_id = None
        self.type = None
        self.x = None
        self.y = None
        self.df = None
        self.zid = None

    def __str__(self):
        return (
            "test id: {} \n"
            "type: {} \n"
            "(x,y): ({:.2f},{:.2f}) \n"
            "First rows of dataframe: \n {}".format(
                self.test_id, self.type, self.x, self.y, self.df.head(2)
            )
        )



from setuptools import setup

exec(open("pygef/_version.py").read())

setup(
    name="pygef",
    version=__version__,
    author="Ritchie Vink",
    author_email="ritchie46@gmail.com",
    url="https://www.ritchievink.com",
    license="mit",
    packages=["pygef", "pygef.been_jefferies", "pygef.robertson"],
    install_requires=[
        "polars>= 0.9.5",
        "matplotlib>= 3.4.2",
        "lxml==4.6.4",
    ],
    python_requires=">=3.7",
    include_package_data=True,
)



import logging
from typing import Union

import pygef.plot_utils as plot
from pygef import been_jefferies, robertson
from pygef.base import Base
from pygef.broxml import _BroXmlCpt
from pygef.gef import _GefCpt

logger = logging.getLogger(__name__)


class Cpt(Base):
    """
    ** Cpt attributes:**
        *Always present:*
            type: str
                Type of the gef file
            project_id: str
                Project id
            x: float
                X coordinate respect to the coordinate system
            y: float
                Y coordinate respect to the coordinate system
            zid: float
                Z coordinate respect to the height system
            height_system: float
                Type of coordinate system, 31000 is NAP
            file_date: datatime.datetime
                Start date time
            test_id: str
                Identifying name of gef file.
            s: str
                String version of gef file.
            df: polars.DataFrame
                DataFrame containing the same column contained in the original .gef file and
                some additional columns [depth, elevation_with_respect_to_nap]

                Tip: Use depth column instead of the penetration_length, the depth is corrected
                with the inclination(if present).

                Note that the Friction ratio is always calculated from the fs and qc values and not parsed from the file.

                If this attribute is called after the classify method the columns relative to the classification
                are also contained.

        *Not always present*

            default: None
            The description is added only for the most important attributes, for the others check:
            https://publicwiki.deltares.nl/download/attachments/102204318/GEF-CPT.pdf?version=1&modificationDate=1409732008000&api=v2

            cpt_class: str
                Cpt class. The format is not standard so it might be not always properly parsed.
            column_void: str
                It is the definition of no value for the gef file
            nom_surface_area_cone_tip: float
                Nom. surface area of cone tip [mm2]
            nom_surface_area_friction_element: float
                Nom. surface area of friction casing [mm2]
            net_surface_area_quotient_of_the_cone_tip: float
                Net surface area quotient of cone tip [-]
            net_surface_area_quotient_of_the_friction_casing: float
                Net surface area quotient of friction casing [-]
            distance_between_cone_and_centre_of_friction_casing: float

            friction_present: float

            ppt_u1_present: float

            ppt_u2_present: float

            ppt_u3_present: float

            inclination_measurement_present: float

            use_of_back_flow_compensator: float

            type_of_cone_penetration_test: float

            pre_excavated_depth: float
                 Pre excavate depth [m]
            groundwater_level: float
                Ground water level [m]
            water_depth_offshore_activities: float
            end_depth_of_penetration_test: float
            stop_criteria: float

            zero_measurement_cone_before_penetration_test: float

            zero_measurement_cone_after_penetration_test: float

            zero_measurement_friction_before_penetration_test: float

            zero_measurement_friction_after_penetration_test: float

            zero_measurement_ppt_u1_before_penetration_test: float

            zero_measurement_ppt_u1_after_penetration_test: float

            zero_measurement_ppt_u2_before_penetration_test: float

            zero_measurement_ppt_u2_after_penetration_test: float

            zero_measurement_ppt_u3_before_penetration_test: float

            zero_measurement_ppt_u3_after_penetration_test: float

            zero_measurement_inclination_before_penetration_test: float

            zero_measurement_inclination_after_penetration_test: float

            zero_measurement_inclination_ns_before_penetration_test: float

            zero_measurement_inclination_ns_after_penetration_test: float

            zero_measurement_inclination_ew_before_penetration_test: float

            zero_measurement_inclination_ew_after_penetration_test : float

            mileage: float
    """

    def __init__(self, path=None, content: dict = None):
        """
        Cpt class.

        Parameters
        ----------
        path:
            Path to the file.
        content: dict
            Dictionary with keys: ["string", "file_type"]
                - string: str
                    String version of the file.
                - file_type: str
                    One of [gef, xml]
        """
        self.net_surface_area_quotient_of_the_cone_tip = None
        self.pre_excavated_depth = None

        super().__init__()

        parsed: Union[_BroXmlCpt, _GefCpt]

        if content is not None:
            assert (
                content["file_type"] == "gef" or content["file_type"] == "xml"
            ), f"file_type can be only one of [gef, xml] "
            assert content["string"] is not None, "content['string'] must be specified"
            if content["file_type"] == "gef":
                parsed = _GefCpt(string=content["string"])
            elif content["file_type"] == "xml":
                parsed = _BroXmlCpt(string=content["string"])

        elif path is not None:
            if path.lower().endswith("gef"):
                parsed = _GefCpt(path)
            elif path.lower().endswith("xml"):
                parsed = _BroXmlCpt(path)
        else:
            raise ValueError("One of [path, (string, file_type)] should be not None.")

        self.__dict__.update(parsed.__dict__)

    def classify(
        self,
        classification,
        water_level_NAP=None,
        water_level_wrt_depth=None,
        p_a=0.1,
        new=True,
    ):
        """
        Classify each row of the cpt type.

        Parameters
        ----------
        classification: str
            Specify the classification, possible choices : "robertson", "been_jefferies".
        water_level_NAP: float, only for cpt type, necessary for the classification: give this or water_level_wrt_depth
            Water level with respect to NAP
        water_level_wrt_depth: float, only for cpt type, necessary for the classification: give this or water_level_NAP
            Water level with respect to the ground_level [0], it should be a negative value.
        p_a: float
            Atmospheric pressure. Default: 0.1 MPa.
        new: bool, default:True
            If True and the classification is robertson, the new(2016) implementation of robertson is used.

        Returns
        -------
        df: polars.DataFrame
            A polars.DataFrame with a classification for each row is returned.

        """
        # todo: refactor arguments, the arguments connected to each other
        #  should be given as a dict or tuple, check order
        water_level_and_zid_NAP = dict(water_level_NAP=water_level_NAP, zid=self.zid)

        if water_level_NAP is None and water_level_wrt_depth is None:
            water_level_wrt_depth = -1
            logger.warning(
                f"You did not input the water level, a default value of -1 m respect to the ground is used."
                f" Change it using the kwagr water_level_NAP or water_level_wrt_depth."
            )

        if classification == "robertson":
            df = robertson.classify(
                self.df,
                water_level_and_zid_NAP=water_level_and_zid_NAP,
                water_level_wrt_depth=water_level_wrt_depth,
                new=new,
                area_quotient_cone_tip=self.net_surface_area_quotient_of_the_cone_tip,
                pre_excavated_depth=self.pre_excavated_depth,
                p_a=p_a,
            )
            return df

        elif classification == "been_jefferies":
            df = been_jefferies.classify(
                self.df,
                water_level_and_zid_NAP=water_level_and_zid_NAP,
                water_level_wrt_depth=water_level_wrt_depth,
                area_quotient_cone_tip=self.net_surface_area_quotient_of_the_cone_tip,
                pre_excavated_depth=self.pre_excavated_depth,
            )
            return df
        else:
            raise ValueError(
                f"Could not find {classification}. Check the spelling or classification not defined in the library"
            )

    def plot(
        self,
        classification=None,
        water_level_NAP=None,
        water_level_wrt_depth=None,
        min_thickness=None,
        p_a=0.1,
        new=True,
        show=False,
        figsize=(11, 8),
        df_group=None,
        grid_step_x=None,
        dpi=100,
        colors=None,
        z_NAP=False,
    ):
        """
        Plot the cpt file and return matplotlib.pyplot.figure .

        Parameters
        ----------
        classification: str, only for cpt type
            If classification ("robertson", "been_jefferies") is specified a subplot is added with the classification
            for each cpt row.
        water_level_NAP: float, only for cpt type, necessary for the classification: give this or water_level_wrt_depth
            Water level with respect to NAP
        water_level_wrt_depth: float, only for cpt type, necessary for the classification: give this or water_level_NAP
            Water level with respect to the ground_level [0], it should be a negative value.
        min_thickness: float, only for cpt type, optional for the classification [m]
            If specified together with the do_grouping set to True, a group classification is added to the plot.
            The grouping is a simple algorithm that merge all the layers < min_thickness with the last above one >
            min_thickness.
            In order to not make a big error do not use a value bigger then 0.2 m
        p_a: float, only for cpt type, optional for the classification
            Atmospheric pressure. Default: 0.1 MPa.
        new: bool, only for cpt type, optional for the classification default:True
            If True and the classification is robertson, the new(2016) implementation of robertson is used.
        show: bool
            If True the plot is showed, else the matplotlib.pytplot.figure is returned
        figsize: tuple
            Figsize of the plot, default (11, 8).
        df_group: polars.DataFrame, only for cpt type, optional for the classification
            Use this argument to plot a defined soil layering next to the other subplots.
            It should contain the columns:
                - layer
                    Name of layer, should be either BeenJefferies of Robertson soil type,
                    if it is different then also the argument colors should be passed.
                - z_centr_NAP
                    Z value of the middle of the layer
                - thickness
                    Thickness of the layer
        grid_step_x: float, only for cpt type, default: None
            Grid step for qc and Fr subplots.
        dpi: int
            Dpi figure
        colors: dict
            Dictionary containing the colors associated to each soil type, if specified
        z_NAP: bool
            If True the Z-axis is with respect to NAP.
        Returns
        -------
        matplotlib.pyplot.figure
        """
        # todo: refactor arguments, the arguments connected to each other should
        #  be given as a dict or tuple, check order
        if classification is None:
            df = self.df
        else:
            df = self.classify(
                classification=classification,
                water_level_NAP=water_level_NAP,
                water_level_wrt_depth=water_level_wrt_depth,
                p_a=p_a,
                new=new,
            )

        return plot.plot_cpt(
            df,
            df_group,
            classification,
            show=show,
            figsize=figsize,
            grid_step_x=grid_step_x,
            colors=colors,
            dpi=dpi,
            z_NAP=z_NAP,
        )
